import numpy as np
import io


def array_to_bytes(dat):
    output = io.BytesIO()
    np.save(output, dat)
    return output.getvalue()


def bytes_to_array(buf):
    return np.load(io.BytesIO(buf))


def testf():
    print("hello")



import numpy as np
from glob import glob
import os
import h5py
from datetime import timedelta, datetime


class H5Stream(object):
    """docstring for H5TTable"""

    def __init__(self, path, dset_key='data'):
        self.path = path
        self.hf = h5py.File(path, 'r')
        self.keys = list(self.hf.keys())
        self.set_dataset(dset_key)
        self.samplerate = self.hf.attrs['samplerate']
        self.channels = self.hf['channels'][:].astype(str)
        # self.stations = self.hf['stations'][:].astype('U4')
        self._nameixs = dict(zip(self.channels, np.arange(len(self.channels))))
        # self.starttime = UTCDateTime(self.hf.attrs['starttime'])
        self.starttime = datetime.fromtimestamp(self.hf.attrs['starttime'])
        self.endtime = self.starttime + timedelta(seconds=self.shape[1] / self.samplerate)
        # self.locations = self.hf['locations'][:]
        # self.coords = self.hf['grid_locs'][:]

    def set_dataset(self, key):
        if key in self.keys:
            self.dset = self.hf[key]
            self.shape = self.dset.shape
        else:
            raise KeyError('dataset %s does not exist' % key)

    def get_row_indexes(self, names):
        if isinstance(names, (list, np.ndarray)):
            return np.array([self._nameixs[name] for name in names])
        else:
            return self._nameixs[names]

    def query(self, channels, t0, t1):
        i0 = self.time_to_index(t0)
        i1 = self.time_to_index(t1)
        irows = self.get_row_indexes(channels)
        # nsamp = int((i1 - i0) * len(irows))
        # print(f"loading {i0}:{i1} for {len(irows)} rows ({nsamp} samples)")

        return self.dset[irows, i0:i1]

    def time_to_index(self, t):
        return int((t - self.starttime).total_seconds() * self.samplerate + 0.5)

    def slice_gen(self, starttime, endtime, chans, wlen, stepsize=None):

        if stepsize is None:
            stepsize = wlen

        wlen_td = timedelta(seconds=wlen)
        stepsize_td = timedelta(seconds=stepsize)
        t0 = starttime

        while (t0 <= endtime - wlen_td):
            t1 = t0 + stepsize_td
            print(t0, t1)
            dat = self.query(chans, t0, t1)
            t0 = t1
            yield dat

    def close(self):
        self.hf.close()



import os
from datetime import datetime, timedelta
import numpy as np
import itertools
from xseis2 import xutil
from xseis2 import xchange
# from xseis2 import xio
from sqlalchemy import MetaData
from xseis2.xsql import Channel, Station, StationPair, StationDvv, XCorr, ChanPair

from loguru import logger


def measure_dvv_stations(session):

    ckeys = unique_ordered(session, XCorr.corr_key)

    stas = []
    for ck in ckeys:
        # stas.extend([c[:-1] for c in ck.split('_')])
        stas.extend([c.split('.')[1] for c in ck.split('_')])

    stas = np.unique(stas)
    times = np.array(session.query(XCorr.start_time).filter_by(status=2).distinct().all()).flatten()

    logger.info(f"processing {len(times)} times for {len(stas)} stations")

    for i, dtime in enumerate(times):
        logger.info(f"time {i} / {len(times)}")

        for sta in stas:
            dkey = f"{str(dtime)} {sta}"
            exists = session.query(StationDvv).filter_by(id=dkey).scalar() is not None

            if exists:
                print(f"{dkey} already exists, skipping")
                continue

            xcorrs = session.query(XCorr).filter(XCorr.start_time == dtime).filter(XCorr.corr_key.like(f"%.{sta}.%")).filter(XCorr.velocity_change.isnot(None)).all()
            if len(xcorrs) == 0:
                continue
            out = np.array([xc.velocity_change for xc in xcorrs])
            length = xcorrs[0].length
            # out = session.query(XCorr.velocity_change).filter(XCorr.start_time == dtime).filter(XCorr.corr_key.like(f"%.{sta}.%")).filter(XCorr.velocity_change.isnot(None)).all()

            # print(sta, len(out))
            session.add(StationDvv(id=dkey, start_time=dtime, station=sta, velocity_change=np.median(out), error=np.std(out), navg=len(out), length=length))
        session.commit()


def measure_dvv_xcorrs(session, rhandle, params):

    ckeys_db = np.array(session.query(XCorr.corr_key).distinct().order_by(XCorr.corr_key.asc()).all()).flatten()
    # ckeys_db = np.unique(session.query(XCorr.corr_key).all())
    logger.info(f'{len(ckeys_db)} unique xcorr pairs in db')

    coda_start_velocity = params.coda_start_velocity
    coda_end_sec = params.coda_end_sec
    dvv_freq_lims = np.array(params.freq_band_measure)
    dvv_wlen_sec = params.wlen_sec
    step_factor = params.step_factor
    nwin_welch = params.coherence_welch_nwin

    for i, ckey in enumerate(ckeys_db):
        logger.info(f"processing {i} / {len(ckeys_db)}")

        cpair = session.query(ChanPair).filter_by(name=ckey).first()
        if cpair is None:
            logger.info(f"{ckey} missing")
            continue

        xcorrs = session.query(XCorr).filter_by(corr_key=ckey).filter_by(status=2).order_by(XCorr.start_time.asc()).all()

        xcorr_load_waveforms(xcorrs, rhandle)

        dist = cpair.dist
        sr = xcorrs[0].samplerate
        coda_start_sec = dist / coda_start_velocity
        print(f"{ckey}: {dist:.2f}m")

        for icc in range(1, len(xcorrs)):
            xref = xcorrs[icc - 1]
            xcur = xcorrs[icc]
            if xcur.velocity_change is not None:
                continue

            # dvv_out = xchange.dvv(xref_sig, xcur_sig, sr, dvv_wlen_sec, dvv_fband, coda_start_sec, coda_end_sec, dvv_outlier_clip=dvv_outlier_clip)
            # dvv_out = xchange.dvv_phase_no_weights(xref_sig, xcur_sig, sr, dvv_wlen_sec, dvv_freq_lims, coda_start_sec, coda_end_sec, step_factor=step_factor)
            dvv_out = xchange.dvv_phase(xref.waveform, xcur.waveform, sr, dvv_wlen_sec, dvv_freq_lims, coda_start_sec, coda_end_sec, nwin_welch=nwin_welch, step_factor=step_factor)
            xcur.velocity_change = dvv_out['dvv']
            xcur.error = dvv_out['regress'][2]
            xcur.pearson = dvv_out['coeff']

        session.commit()


def unique_ordered(session, sql_class_attr):
    return np.array(session.query(sql_class_attr).distinct().order_by(sql_class_attr.asc()).all()).flatten()


def time_group_and_stack_xcorrs(xcorrs, nhour_stack):

    xgroups, dbins = group_xcorrs_by_time(xcorrs, nhour_stack)
    # [len(x) for x in xgroups]

    xcorrs_stack = []
    for xg, new_start_time in zip(xgroups, dbins):
        xnew = stack_xcorrs(xg, new_start_time, nhour_stack)
        xcorrs_stack.append(xnew)

    return xcorrs_stack


def stack_xcorrs(xcorrs, new_start_time, new_length):
    ex = xcorrs[0]

    sig = np.mean(np.array([x.waveform for x in xcorrs]), axis=0)
    nhour_data = np.sum(np.array([x.nstack / 100 * x.length for x in xcorrs]))
    new_nstack = nhour_data / new_length * 100

    xnew = XCorr(corr_key=ex.corr_key, start_time=new_start_time, length=new_length, nstack=new_nstack, samplerate=ex.samplerate, status=2)
    xnew.waveform = sig

    # change status of chunks to processed
    for xc in xcorrs:
        xc.status = 1

    return xnew


def write_xcorrs(xcorrs, session, rhandle):

    pipe = rhandle.pipeline()
    for xc in xcorrs:
        redis_key = f"{str(xc.start_time)} {xc.corr_key}"
        xc.waveform_redis_key = redis_key
        pipe.set(redis_key, xutil.array_to_bytes(xc.waveform))

    pipe.execute()  # add data to redis
    session.add_all(xcorrs)  # add metadata rows to sql
    session.commit()


def group_xcorrs_by_time(xcorrs, blocklen_hours):

    wlen = timedelta(hours=blocklen_hours)

    dates = np.array([x.start_time for x in xcorrs])
    start = xutil.hour_round_down(np.min(dates))
    stop = xutil.hour_round_up(np.max(dates))
    dbins = xutil.datetime_bins(start, stop, wlen)

    xgroups = [[] for i in range(len(dbins))]

    for xc in xcorrs:
        ind = np.argmin(np.abs(dbins - xc.start_time))
        xgroups[ind].append(xc)

    xgroups = np.array(xgroups)
    ikeep = np.where(np.array([len(x) for x in xgroups]) > 0)[0]

    return xgroups[ikeep], dbins[ikeep]


def load_npz_continuous(fname):
    from obspy import UTCDateTime

    with np.load(fname) as npz:
        sr = float(npz['sr'])
        chan_names = npz['chans']
        starttime = UTCDateTime(str(npz['start_time'])).datetime
        data = npz['data'].astype(np.float32)
        data[data == 0] = -1
    endtime = starttime + timedelta(seconds=data.shape[1] / sr)

    return data, sr, starttime, endtime, chan_names


def load_npz_continuous_meta(fname):
    from obspy import UTCDateTime

    with np.load(fname) as npz:
        sr = float(npz['sr'])
        chan_names = npz['chans']
        starttime = UTCDateTime(str(npz['start_time'])).datetime
        # data = npz['data'].astype(np.float32)
        # data[data == 0] = -1
    endtime = starttime + timedelta(minutes=10)
    data = None
    return data, sr, starttime, endtime, chan_names


def ckeys_from_stapairs(pairs):

    db_corr_keys = []

    for pair in pairs:
        # print(pair)
        sta1, sta2 = pair.station1, pair.station2
        for comp1, comp2 in itertools.product(sta1.channels, sta2.channels):
            chan1 = f".{sta1.name}.{comp1.upper()}"
            chan2 = f".{sta2.name}.{comp2.upper()}"
            corr_key = f"{chan1}_{chan2}"
            db_corr_keys.append(corr_key)

    return np.array(db_corr_keys)


# def fill_tables_sta_chan(stations, session, clear=True):

#     if clear:
#         session.query(StationPair).delete()
#         session.query(Channel).delete()
#         session.query(Station).delete()
#         session.commit()

#     sta_rows = []
#     chan_rows = []

#     for sta in sorted(stations, key=lambda x: x.code):
#         chans = sorted(sta.channels, key=lambda x: x.code)
#         chan_names = []

#         for chan in chans:
#             chan_name = f"{sta.code}.{chan.code}"
#             chan_rows.append(Channel(name=chan_name, component=chan.code, station_name=sta.code, quality=1))

#             chan_names.append(chan_name)

#         db_station = Station(name=sta.code, channels=chan_names, location=sta.loc)
#         sta_rows.append(db_station)

#     session.add_all(sta_rows)
#     session.add_all(chan_rows)
#     session.commit()


def fill_table_stations(stations, session, clear=True):

    if clear:
        session.query(Station).delete()

    rows = []

    for sta in sorted(stations, key=lambda x: x.code):
        chans = sorted([c.code for c in sta.channels])
        db_station = Station(name=sta.code, channels=chans, location=sta.loc)
        rows.append(db_station)

    session.add_all(rows)
    session.commit()


def fill_table_station_pairs(session, stations=None, clear=True):
    if stations is None:
        stations = session.query(Station).all()

    if clear:
        session.query(StationPair).delete()

    stations_sorted = sorted(stations, key=lambda x: x.name)

    rows = []
    for sta1, sta2 in itertools.combinations(stations_sorted, 2):
        inter_dist = xutil.dist(sta1.location, sta2.location)
        name = f"{sta1.name}_{sta2.name}"
        pair = StationPair(name=name, dist=inter_dist, station1_name=sta1.name, station2_name=sta2.name)
        rows.append(pair)

    session.add_all(rows)
    session.commit()


def fill_table_xcorrs(dc, sr, nstack, ckeys, starttime, endtime, session, rhandle, nstack_min_percent=0):
    """
    Fill database with cross-correlations
    """
    ncc, nsamp = dc.shape
    logger.info(f'Adding {ncc} correlations for {starttime} to database')
    length_hours = (endtime - starttime).total_seconds() / 3600.

    pipe = rhandle.pipeline()
    rows = []
    for i, sig in enumerate(dc):
        if nstack[i] < nstack_min_percent:
            continue
        ckey = ckeys[i]
        dkey = f"{str(starttime)} {ckey}"

        pipe.set(dkey, xutil.array_to_bytes(sig))
        rows.append(XCorr(corr_key=ckey, start_time=starttime, length=length_hours, nstack=float(nstack[i]), waveform_redis_key=dkey, samplerate=sr, status=0))

    pipe.execute()  # add data to redis
    session.add_all(rows)  # add metadata rows to sql
    session.commit()


def fill_table_chanpairs(session, pair_dist_min=0, pair_dist_max=99999, bad_chans=None):

    logger.info(f'Clear and fill ChanPair PSQL table')
    session.query(ChanPair).delete()

    sta_pairs = session.query(StationPair).filter(StationPair.dist.between(pair_dist_min, pair_dist_max)).filter().all()

    rows = []

    for pair in sta_pairs:
        sta1, sta2 = pair.station1, pair.station2
        dist = pair.dist
        for comp1, comp2 in itertools.product(sta1.channels, sta2.channels):
            chan1 = f".{sta1.name}.{comp1.upper()}"
            chan2 = f".{sta2.name}.{comp2.upper()}"
            corr_key = f"{chan1}_{chan2}"
            if chan1 in bad_chans or chan2 in bad_chans:
                # print(f"bad {corr_key}")
                continue

            # print(corr_key)
            rows.append(ChanPair(name=corr_key, dist=dist, stationpair_name=pair.name))

    session.add_all(rows)  # add rows to sql
    session.commit()

    logger.info(f'Added {len(rows)} channel pairs')


def xcorr_load_waveforms(xcorr_objects, rhandle):
    for xcorr in xcorr_objects:
        xcorr.waveform = xutil.bytes_to_array(rhandle.get(xcorr.waveform_redis_key))


def ckey_dists(ckeys, stations):

    sites = [station.code for station in stations]
    site_locs = [station.loc for station in stations]
    ldict = dict(zip(sites, site_locs))

    dists = dict()
    for ck in ckeys:
        c1, c2 = ck.split('_')
        dd = xutil.dist(ldict[c1[:-2]], ldict[c2[:-2]])
        dists[ck] = dd
    return dists


def sql_drop_tables(db):

    metadata = MetaData(db, reflect=True)

    for table in reversed(metadata.sorted_tables):
        table.drop(db)


def get_continuous_fake(start_time, end_time, stations, fband=[20, 400], sr=1000.0):
    from microquake.core.stream import Trace, Stream
    from obspy import UTCDateTime

    start_time = UTCDateTime(start_time)
    end_time = UTCDateTime(end_time)

    nsamp = int((end_time - start_time) * sr)
    traces = []
    for sta in stations:
        # loc = sta.loc
        dat = xutil.band_noise(fband, sr, nsamp)
        for chan in ['X', 'Y', 'Z']:
            # chans.append(f"{sta.code}.{chan.code}")
            # locs.append(loc)
            tr = Trace(data=dat)
            tr.stats.starttime = start_time
            tr.stats.sampling_rate = sr
            tr.stats.channel = chan
            tr.stats.station = sta
            traces.append(tr)

    st = Stream(traces=traces)
    return st




# def fill_table_xcorrs(hstream, session, rhandle):

#     # channels to correlate can either be (1) taken from db (2) or params
#     chans = hstream.channels
#     sr_raw = hstream.samplerate
#     sr = sr_raw

#     # PARAMS - config
#     whiten_freqs = np.array([60, 80, 320, 350])
#     cclen = 20.0
#     keeplag = 1.0
#     stepsize = cclen
#     onebit = True
#     stacklen = 3600.0
#     # stacklen = 1000

#     # Params to function
#     times = [hstream.starttime + timedelta(seconds=i * stacklen) for i in range(5)]

#     for t0 in times:
#         print(t0)
#         t1 = t0 + timedelta(seconds=stacklen)

#         # python generator which yields slices of data
#         datgen = hstream.slice_gen(t0, t1, chans, cclen, stepsize=stepsize)

#         dc, ckeys_ix = xchange.xcorr_stack_slices(
#             datgen, chans, cclen, sr_raw, sr, keeplag, whiten_freqs, onebit=onebit)
#         ckeys = [f"{ck[0]}_{ck[1]}" for ck in chans[ckeys_ix]]

#         pipe = rhandle.pipeline()
#         rows = []
#         for i, sig in enumerate(dc):
#             # print(i)
#             ckey = ckeys[i]
#             dkey = f"{str(t0)} {ckey}"
#             pipe.set(dkey, xutil.array_to_bytes(sig))
#             rows.append(XCorr(time=t0, ckey=ckey, data=dkey))

#         pipe.execute()  # add data to redis
#         session.add_all(rows)  # add rows to sql
#         session.commit()

# def fill_table_chanpairs(stations, session):

#     chans = []

#     for sta in stations:
#         for chan in sta.channels:
#             chans.append(f".{sta.code}.{chan.code.upper()}")
#     chans = np.array(chans)

#     ckeys = xutil.unique_pairs(chans)
#     ckeys = xutil.ckeys_remove_intersta_str(ckeys)
#     ckeys = np.array([f"{ck[0]}_{ck[1]}" for ck in ckeys])

#     sites = [station.code for station in stations]
#     site_locs = [station.loc for station in stations]
#     ldict = dict(zip(sites, site_locs))

#     rows = []
#     # dists = dict()
#     for ck in ckeys:
#         c1, c2 = ck.split('_')
#         sta1, sta2 = c1[:-2], c1[:-2]
#         dd = xutil.dist(ldict[sta1], ldict[sta2])
#         rows.append(ChanPair(corr_key=ck, inter_dist=dd, station1=sta1, station2=sta2))
#         # dists[ck] = dd

#     # for k, v in dists.items():
#     #     rows.append(ChanPair(corr_key=k, dist_meters=v))
#     #     # print(k, v)

#     session.add_all(rows)  # add rows to sql
#     session.commit()



import numpy as np
from scipy import interpolate
import logging

import scipy.fftpack
import scipy.optimize
import scipy.signal
# from scipy.stats import scoreatpercentile
# from scipy.fftpack.helper import next_fast_len
from xseis2 import xutil
from numpy.polynomial.polynomial import polyfit
import matplotlib.pyplot as plt
from datetime import timedelta, datetime
from loguru import logger
from numba import njit


def ot_keep_sites():

    sites = np.array([18,  20,  22,  24,  26,  28,  30,  32,  41,  44,  46,  48,  50,
                52,  54,  56,  58,  63,  65,  67,  69,  71,  79,  81,  83,  86,
                88,  90, 104, 106, 108, 110, 112, 115, 117, 119, 121, 126, 131])
    return sites


def ot_best_pairs():
    return np.array(['.110.X_.41.X', '.110.Y_.41.X', '.110.Y_.41.Y', '.110.Z_.41.X',
       '.110.Z_.41.Y', '.121.X_.41.X', '.121.X_.41.Y', '.121.Y_.41.X',
       '.121.Y_.41.Y', '.121.Z_.41.X', '.121.Z_.41.Y', '.24.X_.41.X',
       '.24.X_.41.Y', '.24.Y_.41.X', '.24.Y_.41.Y', '.117.X_.119.X', '.117.X_.119.Y', '.117.Z_.119.Y', '.117.Z_.119.Z', '.50.X_.52.Y', '.50.X_.52.Z', '.50.Y_.52.Y', '.50.Y_.52.Z'],
      dtype='<U13')


def ot_bad_chans():
     return np.array(['.131.X', '.131.Y', '.131.Z', '.56.Y', '.67.Y', '.83.X', '.86.X',
       '.86.Y', '.86.Z', '.88.X', '.88.Y', '.88.Z', '.90.X', '.90.Z'],
      dtype='<U6')


def signal_quality(sig):

    npts = len(sig)
    nans = len(np.where(sig == np.nan)[0])
    zeros = len(np.where(sig == 0)[0])
    pos_ratio = 100 * len(np.where(sig > 0)[0]) / npts
    zeros_per = zeros / npts * 100
    nans_per = nans / npts * 100

    print(f"npts: {npts} | zeros: {zeros} ({zeros_per:.2f}%) | nans: {nans} ({nans_per:.2f}%) | npositive {pos_ratio:.3f}%")

    return {'zeros': zeros_per, 'nans': nans_per, 'npos': pos_ratio}


def cc_noise_level(dat):
    c1 = np.mean(dat ** 2, axis=0)
    c2 = np.mean(dat, axis=0) ** 2
    noise = np.sqrt((c1 - c2) / (dat.shape[0] - 1))

    return noise


def assess_snr(xcorrs_dat, nsmooth):
    stack = np.mean(xcorrs_dat, axis=0)
    signal = xutil.smooth(xutil.envelope(stack), nsmooth)

    noise = cc_noise_level(xcorrs_dat)
    noise = xutil.smooth(noise, nsmooth)

    snr = signal / noise

    return snr, signal, noise


def mock_corrs_dvv(nsamp, sr, tt_change_percent, fband_sig, fband_noise, noise_scale):

    keeplag = nsamp // 2
    sig1 = xutil.noise1d(keeplag, fband_sig, sr, scale=1, taplen=0)
    sig2 = stretch(sig1, sr, tt_change_percent)

    # sig1 = np.concatenate((sig1[::-1][1:], sig1))
    # sig2 = np.concatenate((sig2[::-1][1:], sig2))
    sig1 = np.concatenate((sig1[::-1], sig1))
    sig2 = np.concatenate((sig2[::-1], sig2))
    noise1 = xutil.noise1d(nsamp, fband_noise, sr, scale=noise_scale, taplen=0)
    noise2 = xutil.noise1d(nsamp, fband_noise, sr, scale=noise_scale, taplen=0)
    sig1 += noise1
    sig2 += noise2
    return sig1, sig2


def stretch_and_mirror(half_in, sr, tt_change_percent, fband_noise, noise_scale):

    # keeplag = nsamp // 2
    # sig1 = xutil.noise1d(keeplag, fband_sig, sr, scale=1, taplen=0)
    half = stretch(half_in, sr, tt_change_percent)
    full = np.concatenate((half[::-1], half))
    if noise_scale > 0:
        full += xutil.noise1d(len(full), fband_noise, sr, scale=noise_scale, taplen=0)
    return full, half


def linear_regression_zforce(xv, yv, weights):
    out = np.linalg.lstsq(xv[:, None] * weights[:, None], yv * weights, rcond=None)
    # print(out)
    slope = out[0][0]
    yint = 0
    residual = out[1][0] / len(xv)
    # residual = out[1][0]

    return yint, slope, residual


def dvv_phase(cc1, cc2, sr, wlen_sec, freq_lims, coda_start_sec, coda_end_sec, step_factor=4, nwin_welch=4):

    cclen = len(cc1)
    assert(cclen == len(cc2))

    times = xutil.xcorr_lagtimes(cclen, sr)
    icoda = np.where((np.abs(times) < coda_end_sec) & (np.abs(times) > coda_start_sec))

    coeff = xutil.pearson_coeff(cc1[icoda], cc2[icoda])

    wlen = int(wlen_sec * sr)
    coda_start = int(coda_start_sec * sr)
    coda_end = int(coda_end_sec * sr)
    hl = len(cc1) // 2
    iwin = [hl - coda_end, hl + coda_end]

    stepsize = wlen // step_factor
    slices = xutil.build_slice_inds(iwin[0], iwin[1], wlen, stepsize=stepsize)
    fdat1, freqs = fft_slices(cc1, slices, sr)
    fdat2, _ = fft_slices(cc2, slices, sr)
    nwin, nfreq = fdat1.shape

    ix_fkeep = np.where((freqs > freq_lims[0]) & (freqs < freq_lims[1]))[0]
    fkeep = freqs[ix_fkeep] * 2 * np.pi

    welch_groups = xutil.build_welch_wins(0, nwin, nwin_welch, stepsize=1)

    n_measure = len(welch_groups)
    out = np.zeros((n_measure, 2), dtype=np.float32)
    # breakpoint()

    cohs = []
    mcoh = []
    for i, (i0, imid, i1) in enumerate(welch_groups):
        # print(i)
        # coh
        fg1 = fdat1[i0:i1]
        fg2 = fdat2[i0:i1]
        gxy = np.mean(np.conj(fg1) * fg2, axis=0)
        gxx = np.mean(np.conj(fg1) * fg1, axis=0)
        gyy = np.mean(np.conj(fg2) * fg2, axis=0)
        coh = np.abs(gxy) ** 2 / (np.abs(gxx * gyy))
        mcoh.append(np.mean(coh[ix_fkeep]))
        # weights = coh[ix_fkeep] ** 2
        weights = 1.0 / (1.0 / (coh[ix_fkeep] ** 2) - 0.9999)
        weights /= np.sum(weights)
        cohs.append(coh)
        # cohs.append(weights)
        # if i == 98:
            # breakpoint()

        # dvv
        ccf = np.conj(fdat1[imid]) * fdat2[imid]
        phi = np.angle(ccf)
        phi_keep = phi[ix_fkeep]
        regress = linear_regression_zforce(fkeep, phi_keep, weights=weights)
        yint, slope, res = regress
        # res = np.sum((phi_keep - slope * fkeep) ** 2) / (np.size(fkeep) - 1)
        res = np.sum((np.sqrt(weights) * (phi_keep - slope * fkeep)) ** 2)
        out[i] = [slope, res]

        # m=slope, phi=phi_keep, v=fkeep 2pi, w=weights
        # e = np.sum(np.abs(phi_keep - slope * fkeep)) / (np.size(fkeep) - 1)
        # s2x2 = np.sum(fkeep ** 2 * weights ** 2)
        # sx2 = np.sum(weights * fkeep ** 2)
        # e = np.sqrt(e * s2x2 / sx2 ** 2)

        # out[i] = [slope, res / len(fkeep)]
        # out[i] = [slope, res]

    cohs = np.array(cohs)
    mcoh = np.array(mcoh)
    dtt, errors = out.T
    # coh = np.mean(cohs, axis=1)
    res = errors.copy()
    min_res = 0
    res[res < min_res] = min_res
    weights = 1 / res
    # weights = np.ones(len(errors))

    imax = dtt * sr

    # coh = 1 + 1 / (errors + 0.1)
    # coh = 0.5 + 1 / (errors + 0.05)
    # coh = errors
    # w = 1.0 / (1.0 / (coh[index_range] ** 2) - 1.0)
        # w[coh[index_range] >= 0.99] = 1.0 / (1.0 / 0.9801 - 1.0)
    # coh = np.ones(len(imax))

    xv = np.mean(slices[welch_groups[:, 1]], axis=1) - hl

    ikeep = np.arange(len(xv))
    is_coda = np.abs(xv) > coda_start
    n_outlier = 0
    ikeep = np.where((is_coda))[0]

    if n_outlier < 90:
        regress = linear_regression_zforce(xv[ikeep], imax[ikeep], weights[ikeep])
    else:
        regress = [0, 0, 0]

    yint, slope, res = regress
    dvv_percentage = -slope * 100

    print(f"[dvv: {dvv_percentage:.4f}%] [corr_coeff: {coeff:.2f}] [outliers: {n_outlier}%] [res_fit: {res:.4f}]")

    out = {"dvv": float(dvv_percentage), "regress": regress, "xvals": xv, "imax": imax, "coh": weights, "coda_win": [coda_start, coda_end], "ikeep": ikeep, "coeff": float(coeff), "n_outlier": float(n_outlier), "cohs": cohs, "residuals": errors, "freqs": freqs, 'sr': sr, 'mcoh': mcoh}

    return out


def dvv_phase_no_weights(cc1, cc2, sr, wlen_sec, freq_lims, coda_start_sec, coda_end_sec, step_factor=4):

    assert(len(cc1) == len(cc2))
    coeff = xutil.pearson_coeff(cc1, cc2)

    wlen = int(wlen_sec * sr)
    coda_start = int(coda_start_sec * sr)
    coda_end = int(coda_end_sec * sr)

    hl = len(cc1) // 2
    iwin = [hl - coda_end, hl + coda_end]

    stepsize = wlen // step_factor
    slices = xutil.build_slice_inds(iwin[0], iwin[1], wlen, stepsize=stepsize)
    dtt, errors = measure_phase_shift_slices(cc1, cc2, slices, sr, freq_lims).T

    imax = dtt * sr

    xv = np.mean(slices, axis=1) - hl
    ikeep = np.arange(len(xv))

    is_coda = np.abs(xv) > coda_start
    n_outlier = 0
    ikeep = np.where((is_coda))[0]

    # xv_time = xv / sr
    # coh = np.ones(len(errors))
    coh = 1 / (errors + 0.5)
    # coh = errors ** 2

    # if dvv_outlier_clip is not None:
    #     is_outlier = np.abs(imax / xv) < (dvv_outlier_clip / 100)
    #     ikeep = np.where((is_coda) & (is_outlier))[0]
    #     n_outlier = int(100 * (1 - (np.sum(is_outlier) / len(is_outlier))))
    #     # print(f"non-outlier: {np.sum(is_outlier) / len(is_outlier) * 100:.2f}%")
    # else:
    #     ikeep = np.where((is_coda))[0]

    if n_outlier < 90:
        regress = linear_regression_zforce(xv[ikeep], imax[ikeep], coh[ikeep])
        # regress = linear_regression_zforce(xv_time[ikeep], dtt[ikeep], coh[ikeep] ** 2)
    else:
        regress = [0, 0, 0]

    yint, slope, res = regress
    dvv_percentage = -slope * 100

    print(f"[dvv: {dvv_percentage:.4f}%] [corr_coeff: {coeff:.2f}] [outliers: {n_outlier}%] [res_fit: {res:.4f}]")

    out = {"dvv": float(dvv_percentage), "regress": regress, "xvals": xv, "imax": imax, "coh": coh, "coda_win": [coda_start, coda_end], "ikeep": ikeep, "coeff": float(coeff), "n_outlier": float(n_outlier), "errors": errors}

    return out


def dvv(cc1, cc2, sr, wlen_sec, cfreqs, coda_start_sec, coda_end_sec, interp_factor=100, dvv_outlier_clip=None, step_factor=4):

    assert(len(cc1) == len(cc2))

    coeff = xutil.pearson_coeff(cc1, cc2)

    wlen = int(wlen_sec * sr)
    coda_start = int(coda_start_sec * sr)
    coda_end = int(coda_end_sec * sr)

    hl = len(cc1) // 2
    iwin = [hl - coda_end, hl + coda_end]

    stepsize = wlen // step_factor
    slices = xutil.build_slice_inds(iwin[0], iwin[1], wlen, stepsize=stepsize)

    fwins1, filt = windowed_fft(cc1, slices, sr, cfreqs)
    fwins2, filt = windowed_fft(cc2, slices, sr, cfreqs)
    imax, coh = measure_shift_fwins_cc(fwins1, fwins2, interp_factor=interp_factor).T

    xv = np.mean(slices, axis=1) - hl
    ikeep = np.arange(len(xv))

    is_coda = np.abs(xv) > coda_start
    n_outlier = 0

    if dvv_outlier_clip is not None:
        is_outlier = np.abs(imax / xv) < (dvv_outlier_clip / 100)
        ikeep = np.where((is_coda) & (is_outlier))[0]
        n_outlier = int(100 * (1 - (np.sum(is_outlier) / len(is_outlier))))
        # print(f"non-outlier: {np.sum(is_outlier) / len(is_outlier) * 100:.2f}%")
    else:
        ikeep = np.where((is_coda))[0]

    if n_outlier < 90:
        regress = linear_regression_zforce(xv[ikeep], imax[ikeep], coh[ikeep] ** 2)
    else:
        regress = [0, 0, 0]

    yint, slope, res = regress
    dvv_percentage = -slope * 100

    print(f"[dvv: {dvv_percentage:.4f}%] [corr_coeff: {coeff:.2f}] [outliers: {n_outlier}%] [res_fit: {res:.4f}]")

    out = {"dvv": float(dvv_percentage), "regress": regress, "xvals": xv, "imax": imax, "coh": coh, "coda_win": [coda_start, coda_end], "ikeep": ikeep, "coeff": float(coeff), "n_outlier": float(n_outlier)}

    return out


def dvv_sym(cc1, cc2, sr, wlen_sec, cfreqs, coda_start_sec, coda_end_sec, interp_factor=100, dvv_outlier_clip=None, step_factor=4):

    assert(len(cc1) == len(cc2))

    coeff = xutil.pearson_coeff(cc1, cc2)

    wlen = int(wlen_sec * sr)
    coda_start = int(coda_start_sec * sr)
    coda_end = int(coda_end_sec * sr)

    iwin = [coda_start, coda_end]
    # iwin = [0, coda_end]

    stepsize = wlen // step_factor
    slices = xutil.build_slice_inds(iwin[0], iwin[1], wlen, stepsize=stepsize)
    xv = np.mean(slices, axis=1)

    fwins1, filt = windowed_fft(cc1, slices, sr, cfreqs)
    fwins2, filt = windowed_fft(cc2, slices, sr, cfreqs)
    imax, coh = measure_shift_fwins_cc(fwins1, fwins2, interp_factor=interp_factor).T

    ikeep = np.arange(len(xv))

    is_coda = xv > coda_start
    n_outlier = 0

    if dvv_outlier_clip is not None:
        is_outlier = np.abs(imax / xv) < (dvv_outlier_clip / 100)
        ikeep = np.where((is_coda) & (is_outlier))[0]
        n_outlier = int(100 * (1 - (np.sum(is_outlier) / len(is_outlier))))
        # print(f"non-outlier: {np.sum(is_outlier) / len(is_outlier) * 100:.2f}%")
    else:
        ikeep = np.where((is_coda))[0]

    if n_outlier < 90:
        regress = linear_regression_zforce(xv[ikeep], imax[ikeep], coh[ikeep] ** 2)
    else:
        regress = [0, 0, 0]

    yint, slope, res = regress
    dvv_percentage = -slope * 100

    print(f"[dvv: {dvv_percentage:.4f}%] [corr_coeff: {coeff:.2f}] [outliers: {n_outlier}%] [res_fit: {res:.4f}]")

    out = {"dvv": float(dvv_percentage), "regress": regress, "xvals": xv, "imax": imax, "coh": coh, "coda_win": [coda_start, coda_end], "ikeep": ikeep, "coeff": float(coeff), "n_outlier": float(n_outlier)}

    return out


def fft_slices(sig, slices, sr, taper_percent=0.2):

    # print("num slices", len(slices))
    wlen_samp = slices[0][1] - slices[0][0]
    pad = int(2 * wlen_samp)
    taper = xutil.taper_window(wlen_samp, taper_percent)
    freqs = np.fft.rfftfreq(pad, 1.0 / sr)
    nfreq = len(freqs)

    fdat = np.zeros((len(slices), nfreq), dtype=np.complex64)

    for i, sl in enumerate(slices):
        fsig = np.fft.rfft(sig[sl[0]:sl[1]] * taper, n=pad)
        fdat[i] = fsig

    return fdat, freqs


def windowed_fft(sig, slices, sr, corner_freqs, taper_percent=0.2):

    # print("num slices", len(slices))
    wlen_samp = slices[0][1] - slices[0][0]
    pad = int(2 * wlen_samp)
    taper = xutil.taper_window(wlen_samp, taper_percent)
    freqs = np.fft.rfftfreq(pad, 1.0 / sr)
    nfreq = len(freqs)

#     fband = [5, 7, 18, 22]
    filt = xutil.freq_window(corner_freqs, pad, sr)
    # norm = 1.0 / xutil.energy(filt) * 2

    fdat = np.zeros((len(slices), nfreq), dtype=np.complex64)

    for i, sl in enumerate(slices):
        fsig = np.fft.rfft(sig[sl[0]:sl[1]] * taper, n=pad)
        fdat[i] = xutil.norm_energy_freq(filt * xutil.phase(fsig))
        # fdat[i] = xutil.norm_energy_freq(fsig)

    return fdat, filt


def measure_shift_fwins_cc(fwins1, fwins2, interp_factor=1):

    nwin, nfreq = fwins1.shape
    pad = nfreq * 2 - 1

    out = np.zeros((nwin, 2), dtype=np.float32)

    for i, (w1, w2) in enumerate(zip(fwins1, fwins2)):
        ccf = np.conj(w1) * w2
        cc = np.fft.irfft(ccf, n=pad * interp_factor)
        cc = np.roll(cc, len(cc) // 2)
        imax = (np.argmax(cc) - len(cc) // 2) / interp_factor
        out[i] = [imax, np.max(cc) * interp_factor]

    return out


def measure_phase_shift_slices(sig1, sig2, slices, sr, freq_lims, taper_percent=0.2):

    # print("num slices", len(slices))
    wlen_samp = slices[0][1] - slices[0][0]
    # nslices = len(slices)
    taper = xutil.taper_window(wlen_samp, taper_percent)

    pad = int(2 * wlen_samp)
    freqs = np.fft.rfftfreq(pad, 1.0 / sr)
    fsr = 1.0 / (freqs[1] - freqs[0])

    ix_fkeep = (np.array(freq_lims) * fsr + 0.5).astype(int)
    fkeep = freqs[ix_fkeep[0]:ix_fkeep[1]] * 2 * np.pi
    weights = np.ones(len(fkeep))

    out = np.zeros((len(slices), 2), dtype=np.float32)

    for i, sl in enumerate(slices):
        fs1 = np.fft.rfft(sig1[sl[0]:sl[1]] * taper, n=pad)
        fs2 = np.fft.rfft(sig2[sl[0]:sl[1]] * taper, n=pad)
        # ccf = np.conj(xutil.phase(fs1)) * xutil.phase(fs2)
        ccf = np.conj(fs1) * fs2
        phi = np.angle(ccf)
        phi_keep = phi[ix_fkeep[0]:ix_fkeep[1]]
        regress = linear_regression_zforce(fkeep, phi_keep, weights=weights)
        yint, slope, res = regress
        # res = np.max(np.fft.irfft(ccf))
        out[i] = [slope, res]

    return out


def measure_shift_fwins_phase(fwins1, fwins2, freqs, flims):

    nwin, nfreq = fwins1.shape
    fsr = 1.0 / (freqs[1] - freqs[0])
    ixf = (flims * fsr + 0.5).astype(int)
    yv = freqs[ixf[0]:ixf[1]] * 2 * np.pi
    weights = np.ones(len(yv))

    out = np.zeros((nwin, 2), dtype=np.float32)

    for i, (w1, w2) in enumerate(zip(fwins1, fwins2)):
        ccf = np.conj(w1) * w2
        phi = np.angle(ccf)
        phi_win = phi[ixf[0]:ixf[1]]
        regress = linear_regression_zforce(yv, phi_win, weights=weights)
        yint, slope, res = regress
        out[i] = [slope, res]

    return out


# def measure_shift_fwins_phase(fwins1, fwins2, freqs, flim):

#     nwin, nfreq = fwins1.shape
#     fsr = 1.0 / (freqs[1] - freqs[0])
#     flims = np.array([80., 280.])
#     ixf = (flims * fsr + 0.5).astype(int)
#     yv = freqs[ixf[0]:ixf[1]] * 2 * np.pi
#     weights = np.ones(len(yv))

#     out = np.zeros((nwin, 2), dtype=np.float32)

#     for i, (w1, w2) in enumerate(zip(fwins1, fwins2)):
#         ccf = np.conj(w1) * w2
#         phi = np.angle(ccf)
#         phi_win = phi[ixf[0]:ixf[1]]
#         regress = linear_regression_zforce(yv, phi_win, weights=weights)
#         yint, slope, res = regress
#         out[i] = [slope, res]

#     return out


def plot_dvv_sym(vals, dvv_true=None):
    import matplotlib.pyplot as plt

    yint, slope, res = vals['regress']
    xv = vals['xvals']
    coh = vals['coh']
    imax = vals['imax']
    ikeep = vals['ikeep']
    coeff = vals['coeff']
    n_outlier = vals['n_outlier']
    coda_start, coda_end = vals['coda_win']
    dvv_meas = vals['dvv']

    xv_fit = np.concatenate(([0], xv))
    tfit = yint + slope * xv_fit
    plt.plot(xv_fit, tfit, label=f'dvv_meas: {dvv_meas:.3f}')

    if dvv_true is not None:
        tfit2 = yint + (dvv_true / 100) * xv_fit
        plt.plot(xv_fit, tfit2, label=f'dvv_true: {dvv_true:.3f}', color='green', linestyle='--')

    plt.scatter(xv[ikeep], imax[ikeep], c=coh[ikeep])
    plt.colorbar()
    mask = np.ones_like(xv, bool)
    mask[ikeep] = False
    plt.scatter(xv[mask], imax[mask], c='red', alpha=0.2, label='ignored')

    title = f"[dvv: {dvv_meas:.4f}%] [corr_coeff: {coeff:.2f}] [outliers: {n_outlier}%] [res_fit: {res:.4f}]"

    plt.title(title)
    plt.axvline(0, linestyle='--')
    alpha = 0.5
    # vel = 3200.
    # direct = dist / vel * sr
    plt.axvline(coda_start, linestyle='--', color='red', alpha=alpha)
    plt.axvline(coda_end, linestyle='--', color='red', alpha=alpha)
    plt.legend()


def plot_dvv_inds(vals, dvv_true=None):
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(15, 8))

    yint, slope, res = vals['regress']
    xv = vals['xvals']
    coh = vals['coh']
    imax = vals['imax']
    ikeep = vals['ikeep']
    coeff = vals['coeff']
    n_outlier = vals['n_outlier']
    coda_start, coda_end = vals['coda_win']
    dvv_meas = vals['dvv']

    tfit = yint + slope * xv
    plt.plot(xv, tfit, label=f'dvv_meas: {dvv_meas:.3f}')

    if dvv_true is not None:
        tfit2 = yint + (dvv_true / 100) * xv
        plt.plot(xv, tfit2, label=f'dvv_true: {dvv_true:.3f}', color='green', linestyle='--')

    plt.scatter(xv[ikeep], imax[ikeep], c=coh[ikeep])
    plt.colorbar()
    mask = np.ones_like(xv, bool)
    mask[ikeep] = False
    plt.scatter(xv[mask], imax[mask], c='red', alpha=0.2, label='ignored')

    title = f"[dvv: {dvv_meas:.4f}%] [corr_coeff: {coeff:.2f}] [outliers: {n_outlier}%] [res_fit: {res:.4f}]"

    plt.title(title)
    plt.axvline(0, linestyle='--')
    alpha = 0.5
    # vel = 3200.
    # direct = dist / vel * sr
    plt.axvline(coda_start, linestyle='--', color='red', alpha=alpha)
    plt.axvline(-coda_start, linestyle='--', color='red', alpha=alpha)
    plt.legend()
    plt.tight_layout()


def plot_dvv(vals, dvv_true=None):
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(15, 8))

    yint, slope, res = vals['regress']
    slope *= 100
    coh = vals['coh']
    ikeep = vals['ikeep']
    coeff = vals['coeff']
    n_outlier = vals['n_outlier']
    dvv_meas = vals['dvv']
    sr = vals['sr']

    imax = vals['imax'] / sr * 100
    coda_start, coda_end = np.array(vals['coda_win']) / sr
    xv = vals['xvals'] / sr

    tfit = yint + slope * xv
    plt.plot(xv, tfit, label=f'dvv_meas: {dvv_meas:.3f}')

    if dvv_true is not None:
        # tfit2 = yint + (dvv_true / 100) * xv
        tfit2 = yint + dvv_true * xv
        plt.plot(xv, tfit2, label=f'dvv_true: {dvv_true:.3f}', color='green', linestyle='--')

    plt.scatter(xv[ikeep], imax[ikeep], c=coh[ikeep], alpha=0.8)
    plt.colorbar()
    mask = np.ones_like(xv, bool)
    mask[ikeep] = False
    plt.scatter(xv[mask], imax[mask], c='red', alpha=0.2, label='ignored')

    title = f"[dvv: {dvv_meas:.4f}%] [corr_coeff: {coeff:.2f}] [outliers: {n_outlier}%] [res_fit: {res:.4f}]"

    plt.title(title)
    plt.axvline(0, linestyle='--')
    alpha = 0.5
    # vel = 3200.
    # direct = dist / vel * sr
    plt.axvline(coda_start, linestyle='--', color='red', alpha=alpha)
    plt.axvline(-coda_start, linestyle='--', color='red', alpha=alpha)
    plt.xlabel("Lag time (seconds)")
    plt.ylabel("dvv (%)")
    plt.legend()
    plt.tight_layout()


def stream_decompose(st, wlen_sec=None, starttime=None):

    if starttime is None:
        starttime = np.min([tr.stats.starttime for tr in st])
    sr = st[0].stats.sampling_rate
    nchan = len(st)

    if wlen_sec is not None:
        npts_fix = int(wlen_sec * sr)
    else:
        npts_fix = int(np.max([len(tr.data) for tr in st]))

    data = np.zeros((nchan, npts_fix), dtype=np.float32)
    chan_names = []

    for i, tr in enumerate(st):
        i0 = int((tr.stats.starttime - starttime) * sr + 0.5)
        sig = tr.data
        slen = min(len(sig), npts_fix - i0)
        data[i, i0: i0 + slen] = sig[:slen]
        chan_names.append(f".{tr.stats.station}.{tr.stats.channel}")

    return data, sr, starttime, np.array(chan_names)


def xcorr_ckeys_stack_slices(rawdat, sr, ckeys, cc_wlen_sec, keeplag_sec, stepsize_sec=None, whiten_freqs=None, onebit=True, pos_ratio_range=[45, 55]):

    ncc = len(ckeys)
    cclen = int(cc_wlen_sec * sr)
    keeplag = int(keeplag_sec * sr)
    stepsize = int(stepsize_sec * sr)

    padlen = int(cc_wlen_sec * sr * 2)
    nfreq = int(padlen // 2 + 1)
    # whiten_freqs = np.array(whiten_freqs)

    logger.info(f'ncc: {ncc}, cclen: {cc_wlen_sec}s, keeplag: {keeplag_sec}s, stepsize: {stepsize_sec} s')
    # print(cclen, padlen, keeplag, stepsize)

    nchan, nsamp = rawdat.shape
    slices = xutil.build_slice_inds(0, nsamp, cclen, stepsize=stepsize)
    nslices = len(slices)

    whiten_win = None
    if whiten_freqs is not None:
        whiten_freqs = np.array(whiten_freqs)
        whiten_win = xutil.freq_window(whiten_freqs, padlen, sr)

    logger.info(f'Computing {ncc} xcorrs for {nslices} slices of {cc_wlen_sec}s each')

    cc_stack_freq = np.zeros((ncc, nfreq), dtype=np.complex64)
    fdat = np.zeros((nchan, nfreq), dtype=np.complex64)
    # quality = np.zeros((nchan, nslices), dtype=np.float32)
    nstack = np.zeros(ncc, dtype=np.float32)
    stack_flag = np.zeros(nchan, dtype=bool)

    for i, sl in enumerate(slices):
        # print(f"stacking slice {i} / {nslices}")
        fdat.fill(0)

        for isig in range(nchan):
            sig = rawdat[isig, sl[0]:sl[1]].copy()
            pos_ratio = 100 * len(np.where(sig > 0)[0]) / len(sig)
            # quality[isig][i] = pos_ratio

            if pos_ratio < pos_ratio_range[0] or pos_ratio > pos_ratio_range[1]:
                # print(f"chan {isig}: pos_ratio {pos_ratio:.1f} not in range, setting to zero")
                stack_flag[isig] = 0
                continue

            stack_flag[isig] = 1

            fsig = np.fft.rfft(sig, n=padlen)

            if whiten_win is not None:
                fsig = whiten_win * xutil.phase(fsig)
            else:
                fsig /= np.sqrt(xutil.energy_freq(fsig))

            if onebit is True:
                sig = np.sign(np.fft.irfft(fsig))
                fsig = np.fft.rfft(sig)
                fsig /= np.sqrt(xutil.energy_freq(fsig))

            fdat[isig] = fsig

        cc_ikeep = np.where(np.sum(stack_flag[ckeys], axis=1) == 2)[0]
        nstack[cc_ikeep] += 1
        xcorr_stack_freq(fdat, ckeys, cc_stack_freq, cc_ikeep)

    keeplag = int(keeplag)
    cc_stack = np.zeros((ncc, keeplag * 2), dtype=np.float32)

    for i in range(ncc):
        cc = np.fft.irfft(cc_stack_freq[i]) / nslices
        cc_stack[i, :keeplag] = cc[-keeplag:]
        cc_stack[i, keeplag:] = cc[:keeplag]

    nstack = nstack * 100 / nslices

    return cc_stack, nstack


@njit
def xcorr_stack_freq(sigs_freq, cc_keys, cc_stack_freq, cc_ikeep):

    for ixc in cc_ikeep:
        k1, k2 = cc_keys[ixc]
        cc_stack_freq[ixc] += np.conj(sigs_freq[k1]) * sigs_freq[k2]


def plot_tt_change(xv, imax, coh, yint, slope, res, ik):

    tfit = yint + slope * xv

    plt.scatter(xv[ik], imax[ik], c=coh[ik])
    plt.colorbar()
    mask = np.ones_like(xv, bool)
    mask[ik] = False
    plt.scatter(xv[mask], imax[mask], c='red', alpha=0.2)
    # plt.scatter(xv[ik], imax[ik], c='red', alpha=0.5)
    plt.plot(xv, tfit)
    # plt.plot(xv, tfit)
    plt.title("tt_change: %.3f%% ss_res: %.3f " % (slope * 100, res))


def measure_shift_cc(sig1, sig2, interp_factor=1, taper_percent=0.2):

    wlen_samp = len(sig1)
    pad = int(2 * wlen_samp)
    taper = xutil.taper_window(wlen_samp, taper_percent)
    # freqs = np.fft.rfftfreq(pad, 1.0 / sr)
    # nfreq = len(freqs)
    # fsr = 1.0 / (freqs[1] - freqs[0])

    fs1 = np.fft.rfft(sig1 * taper, n=pad)
    fs2 = np.fft.rfft(sig2 * taper, n=pad)

    fs1 /= np.sqrt(xutil.energy_freq(fs1))
    fs2 /= np.sqrt(xutil.energy_freq(fs2))

    ccf = np.conj(fs1) * fs2

    cc = np.fft.irfft(ccf, n=pad * interp_factor) * interp_factor
    cc = np.roll(cc, len(cc) // 2)
    imax = (np.argmax(cc) - len(cc) // 2) / interp_factor
    # tmax = imax / sr

    return imax, np.max(cc), cc
    # return cc, imax


def stretch(sig, sr, tt_change_percent):

    npts = len(sig)
    zpad = npts // 2
    npts_pad = npts + zpad

    psig = np.zeros(npts_pad, dtype=sig.dtype)
    psig[: -zpad] = sig

    x = np.arange(0, npts_pad) / sr
    interp = interpolate.interp1d(x, psig)

    sr_new = sr * (1 + tt_change_percent / 100.)
    xnew = np.arange(0, npts) / sr_new
    newsig = interp(xnew)

    return newsig.astype(sig.dtype)


def stretch_sim_ccs(cclen_sec, sr, tt_changes, cfreqs):
    wlen_sec = cclen_sec / 2
    ncc = len(tt_changes)
    nsamp = int(wlen_sec * sr)
    nsamp_cc = int(wlen_sec * sr * 2)
    ref_half = xutil.noise1d(nsamp, cfreqs, sr, scale=1, taplen=0.0)
    ccs = np.zeros((ncc, nsamp_cc), dtype=np.float32)

    for i, tt_change in enumerate(tt_changes):
        tmp = stretch(ref_half, sr, tt_change)
        cc = np.concatenate((tmp[::-1], tmp))
        ccs[i] = cc

    return ccs

# def nextpow2(val):
#     buf = math.ceil(math.log(val) / math.log(2))
#     return int(math.pow(2, buf))


def nextpow2(x):

    return np.ceil(np.log2(np.abs(x)))


def smooth(x, window='boxcar', half_win=3):
    """ some window smoothing """
    # TODO: docsting
    window_len = 2 * half_win + 1
    # extending the data at beginning and at the end
    # to apply the window at the borders
    s = np.r_[x[window_len - 1:0:-1], x, x[-1:-window_len:-1]]
    if window == "boxcar":
        w = scipy.signal.boxcar(window_len).astype(x.dtype)
    else:
        w = scipy.signal.hanning(window_len).astype(x.dtype)
    y = np.convolve(w / w.sum(), s, mode='valid')
    return y[half_win:len(y) - half_win]


def getCoherence(dcs, ds1, ds2):
    # TODO: docsting
    n = len(dcs)
    coh = np.zeros(n).astype('complex')
    valids = np.argwhere(np.logical_and(np.abs(ds1) > 0, np.abs(ds2 > 0)))
    coh[valids] = dcs[valids] / (ds1[valids] * ds2[valids])
    coh[coh > (1.0 + 0j)] = 1.0 + 0j
    return coh


def apply_phase_shift(sig, nshift):
    fsig = np.fft.rfft(sig)
    nfreq = len(fsig)
    fstep = 0.5 / (nfreq - 1)
    factor = -nshift * 2.0 * np.pi * fstep
    vec = np.zeros(nfreq, dtype=np.complex64)
    inds = np.arange(nfreq)
    vec.real = np.cos(inds * factor)
    vec.imag = np.sin(inds * factor)
    return np.fft.irfft(vec * fsig)


def measure_phase_shift(sig1, sig2, sr, flims=None):

    from obspy.signal.invsim import cosine_taper
    from obspy.signal.regression import linear_regression as obspy_linear_regression

    wlen_samp = len(sig1)
    pad = int(2 ** (nextpow2(2 * wlen_samp)))
    taper = cosine_taper(wlen_samp, 0.85)

    sig1 = scipy.signal.detrend(sig1.copy(), type='linear') * taper
    sig2 = scipy.signal.detrend(sig2.copy(), type='linear') * taper

    freqs = np.fft.rfftfreq(pad, 1.0 / sr)
    fsr = 1.0 / (freqs[1] - freqs[0])

    fs1 = np.fft.rfft(sig1, n=pad)
    fs2 = np.fft.rfft(sig2, n=pad)

    ccf = fs1 * np.conj(fs2)

    if flims is None:
        flims = np.array([0, freqs[-1]])
    print(f"freq_lims: {flims}")

    ixf = (flims * fsr + 0.5).astype(int)
    v = freqs[ixf[0]:ixf[1]] * 2 * np.pi

    # Phase:
    phi = np.angle(ccf)
    phi = np.unwrap(phi)
    phi = phi[ixf[0]:ixf[1]]

    cc = np.fft.irfft(ccf)
    cc = np.roll(cc, len(cc) // 2)
    imax = np.argmax(cc) - len(cc) // 2
    tmax = imax / sr

    m, em = obspy_linear_regression(v, phi)

    return m, tmax


def mwcs_msnoise(current, reference, freqmin, freqmax, df, tmin, window_length, step,
         smoothing_half_win=5):
    """
    :type current: :class:`numpy.ndarray`
    :param current: The "Current" timeseries
    :type reference: :class:`numpy.ndarray`
    :param reference: The "Reference" timeseries
    :type freqmin: float
    :param freqmin: The lower frequency bound to compute the dephasing (in Hz)
    :type freqmax: float
    :param freqmax: The higher frequency bound to compute the dephasing (in Hz)
    :type df: float
    :param df: The sampling rate of the input timeseries (in Hz)
    :type tmin: float
    :param tmin: The leftmost time lag (used to compute the "time lags array")
    :type window_length: float
    :param window_length: The moving window length (in seconds)
    :type step: float
    :param step: The step to jump for the moving window (in seconds)
    :type smoothing_half_win: int
    :param smoothing_half_win: If different from 0, defines the half length of
        the smoothing hanning window.

    :rtype: :class:`numpy.ndarray`
    :returns: [time_axis,delta_t,delta_err,delta_mcoh]. time_axis contains the  central times of the windows. The three other columns contain dt, error and
    mean coherence for each window.
    """

    from obspy.signal.invsim import cosine_taper
    from obspy.signal.regression import linear_regression as obspy_linear_regression

    delta_t = []
    delta_err = []
    delta_mcoh = []
    time_axis = []

    window_length_samples = np.int(window_length * df)

    padd = int(2 ** (nextpow2(window_length_samples) + 2))
    # padd = next_fast_len(window_length_samples)
    count = 0
    tp = cosine_taper(window_length_samples, 0.85)
    minind = 0
    maxind = window_length_samples
    while maxind <= len(current):
        cci = current[minind:(minind + window_length_samples)]
        cci = scipy.signal.detrend(cci, type='linear')
        cci *= tp

        cri = reference[minind:(minind + window_length_samples)]
        cri = scipy.signal.detrend(cri, type='linear')
        cri *= tp

        minind += int(step*df)
        maxind += int(step*df)

        fcur = scipy.fftpack.fft(cci, n=padd)[:padd // 2]
        fref = scipy.fftpack.fft(cri, n=padd)[:padd // 2]

        fcur2 = np.real(fcur) ** 2 + np.imag(fcur) ** 2
        fref2 = np.real(fref) ** 2 + np.imag(fref) ** 2

        # Calculate the cross-spectrum
        X = fref * (fcur.conj())
        if smoothing_half_win != 0:
            dcur = np.sqrt(smooth(fcur2, window='hanning',
                                  half_win=smoothing_half_win))
            dref = np.sqrt(smooth(fref2, window='hanning',
                                  half_win=smoothing_half_win))
            X = smooth(X, window='hanning',
                       half_win=smoothing_half_win)
        else:
            dcur = np.sqrt(fcur2)
            dref = np.sqrt(fref2)

        dcs = np.abs(X)

        # Find the values the frequency range of interest
        freq_vec = scipy.fftpack.fftfreq(len(X) * 2, 1. / df)[:padd // 2]
        index_range = np.argwhere(np.logical_and(freq_vec >= freqmin,
                                                 freq_vec <= freqmax))

        # Get Coherence and its mean value
        coh = getCoherence(dcs, dref, dcur)
        mcoh = np.mean(coh[index_range])

        # Get Weights
        w = 1.0 / (1.0 / (coh[index_range] ** 2) - 1.0)
        w[coh[index_range] >= 0.99] = 1.0 / (1.0 / 0.9801 - 1.0)
        w = np.sqrt(w * np.sqrt(dcs[index_range]))
        w = np.real(w)

        # Frequency array:
        v = np.real(freq_vec[index_range]) * 2 * np.pi

        # Phase:
        phi = np.angle(X)
        phi[0] = 0.
        phi = np.unwrap(phi)
        phi = phi[index_range]

        # Calculate the slope with a weighted least square linear regression
        # forced through the origin
        # weights for the WLS must be the variance !
        m, em = obspy_linear_regression(v.flatten(), phi.flatten(), w.flatten())

        delta_t.append(m)

        # print phi.shape, v.shape, w.shape
        # m=slope, phi=phi_keep, v=fkeep 2pi, w=weights
        e = np.sum((phi - m * v) ** 2) / (np.size(v) - 1)
        s2x2 = np.sum(v ** 2 * w ** 2)
        sx2 = np.sum(w * v ** 2)
        e = np.sqrt(e * s2x2 / sx2 ** 2)

        delta_err.append(e)
        delta_mcoh.append(np.real(mcoh))
        time_axis.append(tmin+window_length/2.+count*step)
        count += 1

        del fcur, fref
        del X
        del freq_vec
        del index_range
        del w, v, e, s2x2, sx2, m, em

    if maxind > len(current) + step*df:
        logging.warning("The last window was too small, but was computed")

    return np.array([time_axis, delta_t, delta_err, delta_mcoh])


def mwcs_msnoise_single(cci, cri, freqmin, freqmax, df, smoothing_half_win=5):

    window_length_samples = len(cci)

    # window_length_samples = np.int(window_length * df)

    padd = int(2 ** (nextpow2(window_length_samples) + 2))
    # padd = next_fast_len(window_length_samples)
    # count = 0
    tp = cosine_taper(window_length_samples, 0.85)
    # minind = 0
    # maxind = window_length_samples
    # while maxind <= len(current):
    #     cci = current[minind:(minind + window_length_samples)]
    cci = scipy.signal.detrend(cci, type='linear')
    cci *= tp

    # cri = reference[minind:(minind + window_length_samples)]
    cri = scipy.signal.detrend(cri, type='linear')
    cri *= tp

    # minind += int(step * df)
    # maxind += int(step * df)

    fcur = scipy.fftpack.fft(cci, n=padd)[:padd // 2]
    fref = scipy.fftpack.fft(cri, n=padd)[:padd // 2]

    fcur2 = np.real(fcur) ** 2 + np.imag(fcur) ** 2
    fref2 = np.real(fref) ** 2 + np.imag(fref) ** 2

    # Calculate the cross-spectrum
    X = fref * (fcur.conj())
    if smoothing_half_win != 0:
        dcur = np.sqrt(smooth(fcur2, window='hanning',
                              half_win=smoothing_half_win))
        dref = np.sqrt(smooth(fref2, window='hanning',
                              half_win=smoothing_half_win))
        X = smooth(X, window='hanning',
                   half_win=smoothing_half_win)
    else:
        dcur = np.sqrt(fcur2)
        dref = np.sqrt(fref2)

    dcs = np.abs(X)

    # Find the values the frequency range of interest
    freq_vec = scipy.fftpack.fftfreq(len(X) * 2, 1. / df)[:padd // 2]
    index_range = np.argwhere(np.logical_and(freq_vec >= freqmin,
                                             freq_vec <= freqmax))

    # Get Coherence and its mean value
    coh = getCoherence(dcs, dref, dcur)
    mcoh = np.mean(coh[index_range])

    # Get Weights
    w = 1.0 / (1.0 / (coh[index_range] ** 2) - 1.0)
    w[coh[index_range] >= 0.99] = 1.0 / (1.0 / 0.9801 - 1.0)
    w = np.sqrt(w * np.sqrt(dcs[index_range]))
    w = np.real(w)

    # Frequency array:
    v = np.real(freq_vec[index_range]) * 2 * np.pi

    # Phase:
    phi = np.angle(X)
    phi[0] = 0.
    phi = np.unwrap(phi)
    phi = phi[index_range]

    # Calculate the slope with a weighted least square linear regression
    # forced through the origin
    # weights for the WLS must be the variance !
    m, em = obspy_linear_regression(v.flatten(), phi.flatten(), w.flatten())

    # delta_t.append(m)

    # print phi.shape, v.shape, w.shape
    e = np.sum((phi - m * v) ** 2) / (np.size(v) - 1)
    s2x2 = np.sum(v ** 2 * w ** 2)
    sx2 = np.sum(w * v ** 2)
    e = np.sqrt(e * s2x2 / sx2 ** 2)

    # delta_err.append(e)
    # delta_mcoh.append(np.real(mcoh))
    # time_axis.append(tmin+window_length/2.+count*step)
    # count += 1

    # return np.array([time_axis, delta_t, delta_err, delta_mcoh]).T
    return m, e, coh



# def xcorr_ckeys_stack_slices2(rawdat, sr, ckeys, cc_wlen_sec, keeplag_sec, stepsize_sec=None, whiten_freqs=None, onebit=True, random_amp=1e-9):

#     ncc = len(ckeys)
#     cclen = int(cc_wlen_sec * sr)
#     keeplag = int(keeplag_sec * sr)
#     stepsize = int(stepsize_sec * sr)

#     padlen = int(cc_wlen_sec * sr * 2)
#     nfreq = int(padlen // 2 + 1)
#     whiten_freqs = np.array(whiten_freqs)

#     logger.info(f'ncc: {ncc}, cclen: {cc_wlen_sec}s, keeplag: {keeplag_sec}s, stepsize: {stepsize_sec} s')
#     # print(cclen, padlen, keeplag, stepsize)

#     nchan, nsamp = rawdat.shape
#     slices = xutil.build_slice_inds(0, nsamp, cclen, stepsize=stepsize)

#     whiten_win = None
#     if whiten_freqs is not None:
#         whiten_win = xutil.freq_window(whiten_freqs, padlen, sr)

#     logger.info(f'Computing {ncc} xcorrs for {len(slices)} slices of {cc_wlen_sec}s each')

#     cc_stack_freq = np.zeros((ncc, nfreq), dtype=np.complex64)

#     for i, sl in enumerate(slices):
#         print(f"stacking slice {i} / {len(slices)}")
#         dat = rawdat[:, sl[0]:sl[1]].copy()

#         if onebit is True:
#             dat = xutil.bandpass(dat, whiten_freqs[[0, -1]], sr)
#             dat = np.sign(dat)

#         fdat = np.fft.rfft(dat, axis=1, n=padlen)

#         if whiten_win is not None:
#             for irow in range(fdat.shape[0]):
#                 fdat[irow] = whiten_win * xutil.phase(fdat[irow])
#         else:
#             for irow in range(fdat.shape[0]):
#                 fdat[irow] /= np.sqrt(xutil.energy_freq(fdat[irow]))

#         xcorr_stack_freq(fdat, ckeys, cc_stack_freq)

#     keeplag = int(keeplag)
#     stack = np.zeros((ncc, keeplag * 2), dtype=np.float32)

#     for i in range(ncc):
#         cc = np.fft.irfft(cc_stack_freq[i]) / len(slices)
#         stack[i, :keeplag] = cc[-keeplag:]
#         stack[i, keeplag:] = cc[:keeplag]

#     return stack


# def xcorr_stack_slices_gen(datgen, chans, cclen, sr_raw, sr_dec, keeplag, whiten_freqs, onebit=True):

#     ckeys = xutil.unique_pairs(np.arange(len(chans)))
#     ckeys = xutil.ckeys_remove_intersta(ckeys, chans)

#     ncc = len(ckeys)
#     padlen = int(cclen * sr_dec * 2)
#     nfreq = int(padlen // 2 + 1)
#     whiten_freqs = np.array(whiten_freqs)

#     dec_factor = None
#     if not np.allclose(sr_raw, sr_dec):
#         dec_factor = int(sr_raw / sr_dec)

#     whiten_win = None
#     if whiten_freqs is not None:
#         whiten_win = xutil.freq_window(whiten_freqs, padlen, sr_dec)

#     cc_stack_freq = np.zeros((ncc, nfreq), dtype=np.complex64)

#     nstack = 0
#     for dat in datgen:

#         dat = xutil.bandpass(dat, whiten_freqs[[0, -1]], sr_raw)

#         if dec_factor is not None:
#             dat = dat[:, ::dec_factor]

#         if onebit is True:
#             dat = np.sign(dat)

#         fdat = np.fft.rfft(dat, axis=1, n=padlen)

#         for irow in range(fdat.shape[0]):
#             fdat[irow] = whiten_win * xutil.phase(fdat[irow])

#         for j, ckey in enumerate(ckeys):
#             k1, k2 = ckey
#             # stack[j] = np.fft.irfft(np.conj(fdat[k1]) * fdat[k2])
#             cc_stack_freq[j] += np.conj(fdat[k1]) * fdat[k2]

#         nstack += 1

#     nlag = int(keeplag * sr_dec)
#     stack = np.zeros((ncc, nlag * 2), dtype=np.float32)

#     for i in range(ncc):
#         cc = np.real(np.fft.irfft(cc_stack_freq[i])) / nstack
#         stack[i, :nlag] = cc[-nlag:]
#         stack[i, nlag:] = cc[:nlag]

#     return stack, ckeys


# def xcorr_stack_slices2(hstream, starttime, endtime, chans, cclen, keeplag, whiten_freqs, onebit=True):

#     # slices = xutil.build_slice_inds(i0, i1, cclen, stepsize=stepsize)
#     sr = hstream.samplerate
#     ckeys = xutil.unique_pairs(np.arange(len(chans)))
#     ckeys = xutil.ckeys_remove_intersta(ckeys, chans)

#     ncc = len(ckeys)
#     padlen = int(cclen * sr * 2)
#     nfreq = int(padlen // 2 + 1)

#     whiten_win = None
#     if whiten_freqs is not None:
#         whiten_win = xutil.freq_window(whiten_freqs, padlen, sr)

#     cc_stack_freq = np.zeros((ncc, nfreq), dtype=np.complex64)
#     cclen_td = timedelta(seconds=cclen)
#     t0 = starttime
#     t1 = starttime + cclen_td
#     nstack = 0

#     # sliding window of cclen until endtime
#     while (t1 < endtime - cclen_td):
#         print(t0, t1)
#         dat = hstream.query(chans, t0, t1)

#         if onebit is True:
#             dat = xutil.bandpass(dat, whiten_freqs[[0, -1]], sr)
#             dat = np.sign(dat)

#         fdat = np.fft.rfft(dat, axis=1, n=padlen)

#         for irow in range(fdat.shape[0]):
#             fdat[irow] = whiten_win * xutil.phase(fdat[irow])

#         for j, ckey in enumerate(ckeys):
#             k1, k2 = ckey
#             # stack[j] = np.fft.irfft(np.conj(fdat[k1]) * fdat[k2])
#             cc_stack_freq[j] += np.conj(fdat[k1]) * fdat[k2]

#         t0 += cclen_td
#         t1 += cclen_td
#         nstack += 1

#     nlag = int(keeplag * sr)
#     stack = np.zeros((ncc, nlag * 2), dtype=np.float32)

#     for i in range(ncc):
#         cc = np.real(np.fft.irfft(cc_stack_freq[i])) / nstack
#         stack[i, :nlag] = cc[-nlag:]
#         stack[i, nlag:] = cc[:nlag]

#     return stack, ckeys


# def linear_regression_old(xv, yv, weights, outlier_sd=None):
#     c, stats = polyfit(xv, yv, 1, full=True, w=weights)
#     yint, slope = c
#     residual = stats[0][0] / len(xv)
#     ikeep = np.arange(len(xv))

#     if outlier_sd is not None:
#         residuals = np.abs(yv - (yint + slope * xv))
#         std_res = np.std(residuals)
#         ikeep = np.where(residuals < std_res * outlier_sd)[0]
#         if len(ikeep) != 0:
#             c, stats = polyfit(xv[ikeep], yv[ikeep], 1, full=True, w=weights[ikeep])
#             yint, slope = c
#             residual = stats[0][0] / len(xv)
#             print("stdev residuals: %.3f" % residual)

#     return yint, slope, residual, ikeep


# def linear_regression2(xv, yv, weights, outlier_val=0.0025):
#     ikeep = np.where(np.abs(yv / xv) < outlier_val)[0]
#     c, stats = polyfit(xv[ikeep], yv[ikeep], 1, full=True, w=weights[ikeep])
#     yint, slope = c
#     residual = stats[0][0] / len(xv)

#     return yint, slope, residual, ikeep


# def linear_regression3(xv, yv, weights):
#     XX = np.vstack((xv, np.ones_like(xv))).T
#     # out = np.linalg.lstsq(XX[:, :-1], yv, rcond=None)
#     # out = np.linalg.lstsq(XX, yv, rcond=None)
#     out = np.linalg.lstsq(XX * weights[:, None], yv * weights, rcond=None)
#     print(out)
#     c = out[0]
#     residual = out[1] / len(xv)
#     slope, yint = c

#     return yint, slope, residual


# def linear_regression_yzero(xv, yv, weights):
#     XX = np.vstack((xv, np.ones_like(xv))).T
#     # out = np.linalg.lstsq(XX[:, :-1], yv, rcond=None)
#     out = np.linalg.lstsq(XX, yv, rcond=None)
#     # c, stats =
#     print(out)
#     c = out[0]
#     residual = out[1] / len(xv)

#     # # c, stats = polyfit(xv[ikeep], yv[ikeep], 1, full=True, w=weights[ikeep])
#     # yint, slope = c
#     slope, yint = c
#     # residual = stats[0][0]
#     return yint, slope, residual




import numpy as np
# import math
# from scipy.fftpack import fft, ifft, fftfreq
# import os
# import pickle
import matplotlib.pyplot as plt
from mayavi import mlab
# import matplotlib.gridspec as gridspec
# import subprocess
# import h5py
# import glob
from xseis2 import xutil


def mlab_contour(g, ncont=10, vfmin=0.1, vfmax=0.1, ranges=None, vlims=None, cbar=True):

	if vlims is None:
		gmin, gmax, p2p = xutil.MinMax(g)
		vmin, vmax = gmin + p2p * vfmin, gmax - vfmax * p2p
	else:
		vmin, vmax = vlims
	contours = list(np.linspace(vmin, vmax, ncont))
	src = mlab.pipeline.scalar_field(g)
	# mlab.pipeline.iso_surface(src, contours=contours, opacity=0.3, colormap='viridis', vmin=vmin, vmax=vmax)
	mlab.outline()
	# mlab.axes(line_width=0.5, xlabel='Z', ylabel='Y', zlabel='X', ranges=ranges)
	mlab.axes(line_width=0.5, xlabel='X', ylabel='Y', zlabel='Z', ranges=ranges)
	mlab.pipeline.iso_surface(src, contours=contours[:-1], opacity=0.2, colormap='viridis', vmin=vmin, vmax=vmax)
	mlab.pipeline.iso_surface(src, contours=contours[-1:], opacity=0.8, colormap='viridis', vmin=vmin, vmax=vmax)
	if cbar:
		mlab.colorbar(orientation='vertical')
	return src


# def power(output, gdef, stalocs=None, labels=None, lines=None, lmax=None, title=None):

def power(output, shape, origin, spacing, stalocs=None, labels=None, lines=None, lmax=None, title=None):
	# shape, origin, spacing = gdef[:3], gdef[3:6], gdef[6]
	grid = output.reshape(shape)

	lims = np.zeros((3, 2))
	lims[:, 0] = origin
	lims[:, 1] = origin + shape * spacing
	lims[0] -= lims[0, 0]
	lims[1] -= lims[1, 0]

	fig = mlab.figure(size=(1000, 901))
	ranges = list(lims.flatten())
	src = mlab_contour(grid, ncont=8, vfmin=0.3, vfmax=0.02, ranges=ranges)
	# x, y, z = (sloc - origin) / spacing
	if stalocs is not None:
		x, y, z = (stalocs - origin).T / spacing
		mlab.points3d(x, y, z, color=(1, 0, 0), scale_factor=1.0)
		if labels is not None:
			for i, lbl in enumerate(labels):
				x, y, z = (stalocs[i] - origin) / spacing
				mlab.text3d(x, y, z, str(lbl))

	if lmax is not None:
		x, y, z = (lmax - origin) / spacing
		mlab.points3d(x, y, z, color=(0, 0, 0), scale_factor=2.0)

	lclr = (0.1, 0.1, 0.1)
	if lines is not None:
		for i, line in enumerate(lines):
			x, y, z = (line - origin).T / spacing
			mlab.plot3d(x, y, z, tube_radius=0.2, color=lclr)
	if title is not None:
		mlab.title(title, size=0.5)


def events(locs, vals, stalocs=None, labels=None, lines=None, lmax=None, title=None):

	fig = mlab.figure(size=(1000, 901))
	x, y, z = locs.T
	mlab.points3d(x, y, z, vals, colormap='viridis', scale_factor=4.0)
	mlab.colorbar(orientation='vertical')

	if stalocs is not None:
		x, y, z = stalocs.T
		mlab.points3d(x, y, z, color=(1, 0, 0), scale_factor=20.0, mode='cube')
		mlab.outline()
		mlab.axes(line_width=0.5, xlabel='X', ylabel='Y', zlabel='Z')
		if labels is not None:
			for i, lbl in enumerate(labels):
				x, y, z = stalocs[i]
				mlab.text3d(x, y, z, str(lbl), scale=35.0)


	lclr = (0.1, 0.1, 0.1)
	if lines is not None:
		for i, line in enumerate(lines):
			x, y, z = line
			mlab.plot3d(x, y, z, tube_radius=0.2, color=lclr)
	if title is not None:
		mlab.title(title, size=0.5)


def power_lims(output, gdef, stalocs=None, lmax=None, lines=None, labels=None):

	lims, spacing = gdef[:6], gdef[6]
	shape = (np.diff(lims)[::2] // spacing).astype(int)
	grid = output.reshape(shape)
	origin = lims[::2]

	fig = mlab.figure(size=(1000, 901))
	# fig = mlab.figure()
	ranges = list(lims.flatten())
	# src = mlab_contour(grid, ncont=6, vfmin=0.1, vfmax=0.0, ranges=ranges)
	src = mlab_contour(grid, ncont=6, vfmin=0.1, vfmax=0.01, ranges=ranges)
	# x, y, z = (sloc - origin) / spacing
	if stalocs is not None:
		x, y, z = (stalocs - origin).T / spacing
		mlab.points3d(x, y, z, color=(1, 0, 0), scale_factor=1.0)
		if labels is not None:
			for i, lbl in enumerate(labels):
				x, y, z = (stalocs[i] - origin) / spacing
				mlab.text3d(x, y, z, str(lbl))
				# mlab.plot3d(x, y, z, tube_radius=0.2, color=lclr)

	if lmax is not None:
		x, y, z = (lmax - origin) / spacing
		mlab.points3d(x, y, z, color=(0, 0, 0), scale_factor=2.0)

	# mlab.view(azimuth=-130, elevation=111, distance='auto', focalpoint='auto',   roll=29)
	lclr = (0.1, 0.1, 0.1)
	if lines is not None:
		for i, line in enumerate(lines):
			x, y, z = (line - origin).T / spacing
			mlab.plot3d(x, y, z, tube_radius=0.2, color=lclr)


def power_new(output, gdef, stalocs=None, labels=None, lines=None, lmax=None, title=None, vfmin=0.1, vfmax=0.005):

	lims, spacing = gdef[:6], gdef[6]
	shape = (np.diff(lims)[::2] // spacing).astype(int)
	grid = output.reshape(shape)
	origin = lims[::2]

	fig = mlab.figure(size=(1000, 901))
	ranges = list(lims.flatten())
	src = mlab_contour(grid, ncont=8, vfmin=vfmin, vfmax=vfmax, ranges=ranges)
	# x, y, z = (sloc - origin) / spacing
	if stalocs is not None:
		x, y, z = (stalocs - origin).T / spacing
		mlab.points3d(x, y, z, color=(1, 0, 0), scale_factor=1.0)
		if labels is not None:
			for i, lbl in enumerate(labels):
				x, y, z = (stalocs[i] - origin) / spacing
				mlab.text3d(x, y, z, str(lbl))

	if lmax is not None:
		x, y, z = (lmax - origin) / spacing
		mlab.points3d(x, y, z, color=(0, 0, 0), scale_factor=2.0)

	lclr = (0.1, 0.1, 0.1)
	if lines is not None:
		for i, line in enumerate(lines):
			x, y, z = (line - origin).T / spacing
			mlab.plot3d(x, y, z, tube_radius=0.2, color=lclr)
	if title is not None:
		mlab.title(title, size=0.5)



# -*- coding: utf-8 -*-
# @Author: Philippe Dales
# @Date:   2018-10-02 13:16:23
# @Last Modified by:   Philippe Dales
# @Last Modified time: 2018-10-10 15:00:26
import numpy as np
import struct
from xseis2 import xutil


def stream_to_array(st, t0, npts_fix, taplen=50):
	sr = st[0].stats.sampling_rate
	nsig = len(st)
	data = np.zeros((nsig, npts_fix), dtype=np.float32)
	for i, tr in enumerate(st):
		i0 = int((tr.stats.starttime - t0) * sr + 0.5)
		sig = tr.data - np.mean(tr.data)
		xutil.taper_data(sig, taplen)
		slen = min(len(sig), npts_fix - i0)
		data[i, i0: i0 + slen] = sig[:slen]
	return data


def prep_stream(st):
	for tr in st:
		tr.stats.station = tr.stats.station.zfill(3)
	st.sort()
	# sr = st[0].stats.sampling_rate
	# st.detrend()
	# if sr > dsr:
	# 	st.decimate(int(sr / dsr))


def build_input_data(st, wlen_sec, dsr=None):
	t0 = np.min([tr.stats.starttime for tr in st])
	sr = st[0].stats.sampling_rate
	data = stream_to_array(st, t0, npts_fix=int(wlen_sec * sr))
	# data -= data.mean(axis=1, keepdims=True)
	# data = data - data.mean(axis=1, keepdims=True)
	if dsr is not None and sr > dsr:
		data = xutil.decimate(data, sr, int(sr / dsr))
	stations = np.array([tr.stats.station for tr in st])
	unique = np.unique(stations)
	unique_dict = dict(zip(unique, np.arange(len(unique))))
	chanmap = np.array([unique_dict[chan] for chan in stations], dtype=np.uint16)
	return data, t0, unique, chanmap


def encode_for_kafka(ot_epoch, loc, power):
	msg = np.array([ot_epoch, loc[0], loc[1], loc[2], power], dtype=np.float64)
	kaf_msg = struct.pack('%sd' % len(msg), *msg)
	kaf_key = str(ot_epoch).encode('utf-8')
	kaf_key = ("iloc_%d" % (ot_epoch)).encode('utf-8')

	return kaf_msg, kaf_key


def parse_event_xml(fle):
	import xml.etree.ElementTree as ET
	tree = ET.parse(fle)
	root = tree.getroot()
	vals = root[0][0]
	loc = np.zeros(3)
	loc[0] = vals.find("{MICROQUAKE}LOCATION_X").text
	loc[1] = vals.find("{MICROQUAKE}LOCATION_Y").text
	loc[2] = vals.find("{MICROQUAKE}LOCATION_Z").text

	return loc


def split_mseed(st, wlen):
	from obspy import Stream, Trace
	npts = len(st[0].data)
	nchunks = int(npts / wlen)
	for i in range(nchunks):
		stnew = Stream()
		# stmp = st.copy()
		for tr in st:
			trnew = Trace()
			trnew.stats = tr.stats

	




from sqlalchemy import Column, Integer, String, DateTime, Float, ForeignKey, Boolean
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql import ARRAY

Base = declarative_base()


class Channel(Base):
    __tablename__ = 'channel'

    name = Column(String(10), primary_key=True)
    component = Column(String(5))
    # station = Column(String(10))
    station_name = Column(String(10), ForeignKey('station.name'))
    station = relationship("Station")
    quality = Column(Integer)
    samplerate = Column(Float)

    def __repr__(self):
        return f"<Channel ({self.name}) [samplerate: {self.samplerate} Hz] [quality: {self.quality}]>"


class Station(Base):
    __tablename__ = 'station'

    name = Column(String(10), primary_key=True)
    channels = Column(ARRAY(String(10)))
    location = Column(ARRAY(Float))

    def __repr__(self):
        str_loc = ", ".join([f"{v:.0f}" for v in self.location])
        str_chans = ",".join(self.channels)
        return f"<Station ({self.name}) [{str_chans}] [{str_loc}]>"


class ChanPair(Base):
    __tablename__ = 'chanpair'

    name = Column(String(20), primary_key=True)
    stationpair_name = Column(String(20), ForeignKey('stationpair.name'))
    stationpair = relationship("StationPair", lazy='select')
    dist = Column(Float)

    def __repr__(self):
        return f"<ChanPair ({self.name}, [{self.dist:.2f} m]>"


class StationPair(Base):
    __tablename__ = 'stationpair'

    name = Column(String(20), primary_key=True)
    station1_name = Column(String(10), ForeignKey('station.name'))
    station2_name = Column(String(10), ForeignKey('station.name'))
    station1 = relationship("Station", foreign_keys=[station1_name])
    station2 = relationship("Station", foreign_keys=[station2_name])
    dist = Column(Float)

    def __repr__(self):
        return f"<StationPair ({self.name}) [{self.dist:.2f} m]>"


class XCorr(Base):
    __tablename__ = 'xcorr'

    id = Column(Integer, primary_key=True)
    corr_key = Column(String(20))

    start_time = Column(DateTime)
    length = Column(Float)  # length of stack in hours
    nstack = Column(Float)  # percentage of data kept
    waveform_redis_key = Column(String(50))  # redis key of cross-correlation waveform
    samplerate = Column(Float)

    # status 0: unprocessed chunk  1: processed chunk  2: stacked (used for dvv)
    status = Column(Integer)
    velocity_change = Column(Float)  # dv/v measurement versus reference trace
    error = Column(Float)
    pearson = Column(Float)  # pearson coeff with reference dvv trace (coda only)

    def __repr__(self):

        return f"<Xcorr ({self.corr_key}) {self.start_time} [hours: {self.length:.2f}] [status: {self.status}] [dvv: {self.velocity_change}, error: {self.error}]"


class StationDvv(Base):
    __tablename__ = 'stationdvv'

    id = Column(String(40), primary_key=True)
    # id = Column(Integer, primary_key=True)
    station = Column(String(10))  # seismic station ID
    start_time = Column(DateTime)  # measurement start time
    length = Column(Float)  # length of stack in hours
    # length = Column(String(10))  # identifier of length (e.g 'hourly', 'daily')
    velocity_change = Column(Float)  # dv/v measurement
    error = Column(Float)  # error on the dv/v measurement
    navg = Column(Integer)  # number of chanpair measurements averaged

    def __repr__(self):
        return f"<StationDvv ({self.station}) {self.start_time} [hours: {self.length:.2f}]  [dvv: {self.velocity_change:.4f}, error: {self.error:.2e}, navg: {self.navg}]>"


# class DataFile(Base):
#     __tablename__ = 'datafile'

#     name = Column(String(50), primary_key=True)
#     status = Column(Boolean)

#     def __repr__(self):
#         return f"<DataFile ({self.name}) [processed_status: {self.status}]>"



# GET /api/v1/inventory/noise_correlations
# HTTP 200 OK
# Allow: GET, POST, HEAD, OPTIONS
# Content-Type: application/json
# Vary: Accept

# {
#     "next": null,
#     "previous": null,
#     "cursor_next": null,
#     "cursor_previous": null,
#     "current_page": 1,
#     "total_pages": 1,
#     "count": 1,
#     "results": [
#         {
#             "id": 1,
#             "sensor": {
#                 "id": 1,
#                 "code": "1"
#             },
#             "start_time": "2011-01-01T11:01:11Z",
#             "length": 111.0,
#             "velocity_change": 11.0,
#             "error": 1.0
#         }
#     ]
# }
    # corr_key = Column(String(20), ForeignKey('chanpair.corr_key'))
    # chan_pair = Column(String(20))

    # stationpair_name = Column(String(20))
    # stationpair_name = Column(String(20), ForeignKey('stationpair.name'))
    # stationpair = relationship("StationPair")

    # channel1_name = Column(String(10), ForeignKey('channel.name'))
    # channel2_name = Column(String(10), ForeignKey('channel.name'))
    # channel1 = relationship("Channel", foreign_keys=[channel1_name])
    # channel2 = relationship("Channel", foreign_keys=[channel2_name])



""" utils."""

import numpy as np
import math
from scipy.fftpack import fft, ifft, fftfreq
from scipy.signal import sosfilt, zpk2sos, iirfilter
import os
import glob
import itertools
import io
from scipy.stats import linregress
from datetime import timedelta, datetime
from scipy.signal import hilbert
from scipy.ndimage.filters import gaussian_filter


def hour_round_nearest(t):
    # Rounds to nearest hour by adding a timedelta hour if minute >= 30
    return (t.replace(second=0, microsecond=0, minute=0, hour=t.hour) + timedelta(hours=t.minute // 30))


def hour_round_down(t):
    return t.replace(second=0, microsecond=0, minute=0, hour=t.hour)


def hour_round_up(t):
    return t.replace(second=0, microsecond=0, minute=0, hour=t.hour) + timedelta(hours=1)


def datetime_bins(start, stop, wlen, stepsize=None):

    if stepsize is None:
        stepsize = wlen

    bins = []
    curr = start
    while curr < stop:
        d0 = curr
        d1 = curr + wlen
        # bins.append([d0, d1])
        bins.append(d0)
        curr += stepsize

    return np.array(bins)


def ckeys_remove_chans(ckeys, names):
    ikeep = []
    for i, ck in enumerate(ckeys):
        c1, c2 = ck.split('_')
        if c1 in names or c2 in names:
            continue
        ikeep.append(i)

    return ckeys[np.array(ikeep)]


def ckey_to_chan_stats(ckeys, vals):

    d = dict()
    for i, ck in enumerate(ckeys):
        c1, c2 = ck.split('_')
        if c1 not in d:
            d[c1] = [vals[i]]
            continue
        if c2 not in d:
            d[c2] = [vals[i]]
            continue

        d[c1].append(vals[i])
        d[c2].append(vals[i])

    return np.array(list(d.keys())), np.array(list(d.values()))


def linear_detrend_nan(y):
    x = np.arange(len(y))
    not_nan_ind = ~np.isnan(y)
    m, b, r_val, p_val, std_err = linregress(x[not_nan_ind], y[not_nan_ind])
    detrend_y = y - (m * x + b)
    return detrend_y


def roundup(x, nearest_multiple):
    return int(math.ceil(x / nearest_multiple)) * int(nearest_multiple)


def whiten_sig(sig, sr, whiten_freqs, pad_multiple=None):
    if pad_multiple is None:
        npad = len(sig)
    else:
        npad = roundup(len(sig), pad_multiple)

    whiten_win = freq_window(whiten_freqs, npad, sr)
    fsig = np.fft.rfft(sig, n=npad)
    fsig = whiten_win * phase(fsig)
    return np.fft.irfft(fsig)[:len(sig)]


def symmetric(data):
    """Create symmetric signal."""
    if data.ndim == 1:
        split = split_causals(data)
        split[0][0] = 0
        sym = np.sum(split, axis=0) / 2.
    elif data.ndim == 2:
        sym = (acaus2d(data) + caus2d(data)) / 2

    return sym


def split_causals(sig, overlap=0):
    """Split signals into causal/acausal."""
    length = len(sig)
    lh = length // 2
    pos_start = lh - overlap
    neg_end = lh + overlap

    if length % 2 != 0:
        neg_end += 1

    return np.array([sig[pos_start:], sig[:neg_end][::-1]])


def acaus2d(data):
    return np.fliplr(data[:, :data.shape[1] // 2])


def caus2d(data):
    return data[:, data.shape[1] // 2:]


def average_adjacent_rows(dat, nrow_avg):
    nrow, ncol = dat.shape
    nrow_new = nrow // nrow_avg
    row_mutliple = int(nrow_new * nrow_avg)

    out = dat[:row_mutliple].reshape((nrow_new, nrow_avg, ncol))
    out = np.mean(out, axis=1)
    return out


def to_dict_of_lists(list_of_dicts):
    ex_dict = list_of_dicts[0]
    dict_of_lists = {k: [] for k in ex_dict.keys()}

    for d in list_of_dicts:
        for k, v in d.items():
            dict_of_lists[k].append(v)

    for k, v in dict_of_lists.items():
        dict_of_lists[k] = np.array(v)

    return dict_of_lists


def maxnorm(dat, scale=1):
    out = dat - np.mean(dat)
    return out * scale / np.max(np.abs(out))


def xcorr_lagtimes(nsamp, sr=1):
    hl = nsamp / (2 * sr)
    return np.linspace(-hl, hl, nsamp)


def check_onebit_bool_conversion(sig_raw, set_zero_random=False):

    sig = sig_raw.copy()

    # to bool
    if set_zero_random is True:
        izero = np.where(sig == 0)[0]
        sig[izero] = np.random.choice([-1, 1], len(izero))
    sig[sig < 0] = 0
    sig = np.sign(sig).astype(np.bool_)

    # from bool
    sig = sig.astype(np.float32)
    sig[sig == 0] = -1

    return sig


def index_ckeys_split(corr_keys, chan_names):
    cdict = dict(zip(chan_names, np.arange(len(chan_names))))
    ckeys_ix = []
    for key in corr_keys:
        k0, k1 = key.split('_')
        ckeys_ix.append([cdict[k0], cdict[k1]])

    ckeys_ix = np.array(ckeys_ix)
    return ckeys_ix


def array_to_bytes(dat):
    output = io.BytesIO()
    np.save(output, dat)
    return output.getvalue()


def bytes_to_array(buf):
    return np.load(io.BytesIO(buf))


def acausal(data):
    return np.fliplr(data[:, :data.shape[1] // 2])


def causal(data):
    return data[:, data.shape[1] // 2:]


def ricker(freq, sr, length=0.2):
    f = freq
    dt = 1. / sr
    t = np.arange(-length / 2, (length - dt) / 2, dt)
    y = (1.0 - 2.0*(np.pi**2)*(f**2)*(t**2)) * np.exp(-(np.pi**2)*(f**2)*(t**2))
    return t, y


# def sizemb(nfloat):
#     size = nfloat * 4.0 / 1024**2
#     print("%d float = %.2f mb" % (nfloat, size))


def sizeof(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi']:
        if abs(num) < 1024.0:
            # return "%3.1f%s%s" % (num, unit, suffix)
            return [num, unit + suffix]
        num /= 1024.0


def ttsamp(dist, vel, sr):
    return int(dist / vel * sr + 0.5)


def integrate(sig):
    from scipy.integrate import cumtrapz
    return cumtrapz(sig, initial=0)


def attenuate(sig, sr, dist, Q, vel, gspread=True):

    npts = len(sig)
    fsig = fft(sig)
    freqs = fftfreq(npts, d=1. / sr)
    tstar = dist / (vel * Q)
    factor = np.exp(-np.pi * np.abs(freqs) * tstar)
    fsig *= factor
    sig = np.real(ifft(fsig))
    if gspread:
        sig /= dist
    return sig


def cc_avg_coeff(ccs, ckeys):
    N = np.unique(ckeys).shape[0]
    vals = np.zeros(N)
    count = np.zeros(N)

    for i, k in enumerate(ckeys):
        mx = np.max(ccs[i])
        vals[k] += mx
        count[k] += 1

    return vals / count


def get_pt(index, shape, spacing, origin):
    nx, ny, nz = shape
    iz = index % nz
    iy = ((index - iz) // nz) % ny
    ix = index // (nz * ny)

    loc = np.array([ix, iy, iz]) * spacing + origin
    return loc


def imax_to_xyz_gdef(index, gdef):
    shape, origin, spacing = gdef[:3], gdef[3:6], float(gdef[6])
    nx, ny, nz = shape
    iz = index % nz
    iy = ((index - iz) // nz) % ny
    ix = index // (nz * ny)

    loc = np.array([ix, iy, iz], dtype=float) * spacing + origin
    return loc


def gdef_to_points(gdef):

    shape, origin, spacing = gdef[:3], gdef[3:6], float(gdef[6])
    # nx, ny, nz = shape
    maxes = origin + shape * spacing
    x = np.arange(origin[0], maxes[0], spacing).astype(np.float32)
    y = np.arange(origin[1], maxes[1], spacing).astype(np.float32)
    z = np.arange(origin[2], maxes[2], spacing).astype(np.float32)
    points = np.zeros((np.product(shape), 3), dtype=np.float32)
    # points = np.stack(np.meshgrid(x, y, z), 3).reshape(3, -1).astype(np.float32)
    ix = 0
    for xv in x:
        for yv in y:
            for zv in z:
                points[ix] = [xv, yv, zv]
                ix += 1
    return points


def points2d(x, y):
    return np.transpose([np.tile(x, len(y)), np.repeat(y, len(x))])


def points3d(x, y, z):
    size = len(x) * len(y) * len(z)
    points = np.zeros((size, 3), dtype=np.float32)
    ix = 0
    for xv in x:
        for yv in y:
            for zv in z:
                points[ix] = [xv, yv, zv]
                ix += 1
    return points


def ttable_from_nll_grids(path, key="OT.P"):
    fles = np.sort(glob.glob(os.path.join(path, key + '*.time.buf')))
    hfles = np.sort(glob.glob(os.path.join(path, key + '*.time.hdr')))
    assert(len(fles) == len(hfles))
    stas = np.array([f.split('.')[-3].zfill(3) for f in fles], dtype='S4')
    isort = np.argsort(stas)
    fles = fles[isort]
    hfles = hfles[isort]
    names = stas[isort]

    vals = [read_nll_header(fle) for fle in hfles]
    sloc, shape, org, spacing = vals[0]
    slocs = np.array([v[0] for v in vals], dtype=np.float32)
    ngrid = np.product(shape)

    nsta = len(fles)
    tts = np.zeros((nsta, ngrid), dtype=np.float32)

    for i in range(nsta):
        tts[i] = np.fromfile(fles[i], dtype='f4')

    gdef = np.concatenate((shape, org, [spacing])).astype(np.int32)

    ndict = {}
    for i, sk in enumerate(names):
        ndict[sk.decode('utf-8')] = i

    return tts, slocs, ndict, gdef


def read_nll_header(fle):
    # print(fle)
    dat = open(fle).read().split()
    shape = np.array(dat[:3], dtype=int)
    org = np.array(dat[3:6], dtype=np.float32) * 1000.
    spacing = (np.array(dat[6:9], dtype=np.float32) * 1000.)[0]
    sloc = np.array(dat[12:15], dtype=np.float32) * 1000.

    return sloc, shape, org, spacing


def g2mad(grid):
    # return max, median and median absolute dev
    med = np.median(grid)
    mad = np.median(np.abs(grid - med))
    # mad = max(mad, 1)
    out = (grid - med) / mad
    return out


def randomize_freq_band(sig, band, sr):

    fsig = np.fft.rfft(sig)
    freqmin, freqmax = band
    freqs = np.fft.rfftfreq(len(sig), 1.0 / sr)
    idx = np.where(np.logical_and(freqs >= freqmin, freqs <= freqmax))[0]

    part = fsig[idx]
    amps = np.abs(part)
    rand = np.random.uniform(-np.pi, np.pi, len(part))
    angs = np.exp(1j * rand)
    fsig[idx] = amps * angs

    return np.fft.irfft(fsig)


# def randomize_band_freq(fsig, band, sr):

#     fnew = fsig.copy()
#     freqmin, freqmax = band
#     samples = len(fnew)
#     freqs = np.abs(fftfreq(samples, 1. / sr))
#     idx = np.where(np.logical_and(freqs >= freqmin, freqs <= freqmax))[0]

#     part = fnew[idx]
#     amps = np.abs(part)
#     rand = np.random.uniform(-np.pi, np.pi, len(part))
#     angs = np.exp(1j * rand)
#     fnew[idx] = amps * angs

#     return fnew


def roll_data(data, tts):
    droll = np.zeros_like(data)

    for i, sig in enumerate(data):
        droll[i] = np.roll(sig, -tts[i])
    return droll


def velstack(data, dists2src, sr, vels):

    dnorm = norm2d(data)
    dstack = np.zeros((len(vels), dnorm.shape[1]), dtype=np.float32)
    for ivel, vel in enumerate(vels):
        shifts = (dists2src / vel * sr + 0.5).astype(int)
        for i, shift in enumerate(shifts):
            dstack[ivel] += np.roll(dnorm[i], -shift)
    return dstack


def chan_groups(chanmap):
    return [np.where(sk == chanmap)[0] for sk in np.unique(chanmap)]


def comb_channels(data, cmap):

    groups = [np.where(sk == cmap)[0] for sk in np.unique(cmap)]
    dstack = np.zeros((len(groups), data.shape[1]))

    for i, grp in enumerate(groups):
        dstack[i] = np.mean(np.abs(data[grp]), axis=0)

    return dstack


def mlab_coords(locs, lims, spacing):
    return (locs - lims[:, 0]).T / spacing


def SearchClusters(data, dmin, zmin=0):

    inds = np.arange(data.shape[1])

    slocs = []
    vals = []
    for j, ix in enumerate(inds):
        print(j)
        lmax = data[:-1, ix, 1:]
        tmp = []
        for k, centroid in enumerate(lmax):
            if centroid[2] < zmin:
                tmp.append(0)
                continue
            diff = np.linalg.norm(lmax - centroid, axis=-1)
            tmp.append(np.where(diff < dmin)[0].size)

        imax = np.argmax(tmp)
        vals.append(tmp[imax])
        slocs.append(lmax[imax])

    vals = np.array(vals)
    slocs = np.array(slocs)

    return vals, slocs


def MeanDist(data):

    inds = np.arange(data.shape[1])
    vals = []
    for j, ix in enumerate(inds):
        print(j)
        lmax = data[:-1, ix, 1:]
        x, y, z = lmax.T
        err = np.std(x) + np.std(y) + np.std(z)
        vals.append(err)

    vals = np.array(vals) / 3

    return vals


def shift_locs(locs, unshift=False, vals=np.array([1.79236297e+05, 7.09943400e+06, 2.49199997e+02])):
    vals = np.array(vals)
    locs[:, 2] *= -1
    if unshift is True:
        return locs + vals
    else:
        return locs - vals


def shift_locs_ot(locs, unshift=False, vals=np.array([650000., 4766000., 0]), zdepth=1200.):
    vals = np.array(vals)
    lnew = locs.copy()

    if unshift is True:
        # lnew[:, 2] *= -1
        lnew.T[2] = (lnew.T[2] - zdepth) * -1
        return lnew + vals
    else:
        lnew.T[2] = zdepth - lnew.T[2]
        return lnew - vals

# def shift_locs_ot(locs, unshift=False, vals=np.array([650000., 4766000., 0])):
#   vals = np.array(vals)
#   if unshift is True:
#       return locs + vals
#   else:
#       return locs - vals


def pearson_coeff(sig1, sig2):
    return np.dot(sig1, sig2) / np.sqrt(energy(sig1) * energy(sig2))


def normVec(v):
    return v / norm(v)


def norm(v):
    return math.sqrt(np.dot(v, v))


def angle_between_vecs(v1, v2):
    return math.acos(np.dot(v1, v2) / (norm(v1) * norm(v2)))


def dist(l1, l2):
    return norm(np.array(l1) - np.array(l2))


def dist2many(loc, locs):
    # return norm(l1 - l2)
    return np.linalg.norm(locs - loc, axis=1)


def build_PShSv_matrix(vec):
    """Create orientations for channels to be rotated."""
    P = vec / norm(vec)
    SH = np.array([P[1], -P[0], 0])
    SV = np.cross(SH, P)
    return np.array([P, SH, SV])


def rotation_matrix(axis, theta):
    """Return ccw rotation about the given axis by theta radians."""
    axis = np.asarray(axis)
    theta = np.asarray(theta)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])


def rotate_one(data, orients, uvec):
    """Rotate channel data with respect to uvec."""
    contribs = np.sum(orients * uvec, axis=1)
    return np.sum(data * contribs[np.newaxis, :].T, axis=0)


def apply_rotation(vec, theta, axis, eps=1e-10):
    """Apply rotation_matrix to all vectors."""
    for i in range(3):
        vec[i] = np.dot(rotation_matrix(axis, theta), vec[i])
    vec[np.abs(vec) < eps] = 0


def axes_from_orient(az, dip, roll):
    vec = np.array([[1, 0, 0], [0, 1, 0],  [0, 0, 1]]).astype(float)
    apply_rotation(vec, az, axis=vec[2])
    apply_rotation(vec, dip, axis=vec[1])
    apply_rotation(vec, roll, axis=vec[0])
    return vec


def MedAbsDev(grid):
    return np.median(np.abs(grid - np.median(grid)))


def MinMax(a):
    mn, mx = np.min(a), np.max(a)
    return mn, mx, mx - mn


def read_meta(fname):

    meta = np.loadtxt(fname)
    spacing = meta[6]
    lims = meta[:6].reshape(3, 2)
    shape = (np.diff(lims, axis=1).T[0] // spacing).astype(int)
    # xl, yl, zl = lims
    return lims, spacing, shape


def combine_grids(fles, shape):

    nx, ny, nz = shape

    nfle = len(fles)
    # shape = (nx, nx, nx)
    grids = np.zeros((nfle, nz, ny, nx), dtype=np.float32)

    for i, fn in enumerate(fles):
        # print(fn)
        # xl, yl, zl = lims.reshape(3, 2)
        grid = np.load(fn).reshape(nz, ny, nx)
        grids[i] = grid

    return grids


def combine_grids_nll(fles, shape):

    nx, ny, nz = shape
    nfle = len(fles)
    grids = np.zeros((nfle, nx, ny, nz), dtype=np.float32)

    for i, fn in enumerate(fles):
        grids[i] = np.load(fn).reshape(shape)

    return grids


def xyz_max(grid, lims, spacing):
    # thresh = np.std(grid) * nstd
    iwin = np.argmax(grid)

    pt = np.array(np.unravel_index(iwin, grid.shape))[::-1]
    pt = pt * spacing + lims[:, 0]
    return pt


def xyz_index(loc, lims, spacing):
    # thresh = np.std(grid) * nstd
    shape = (np.diff(lims).T[0] / spacing).astype(int)
    nx, ny, nz = shape
    # x, y, z = loc
    inds = ((loc - lims[:, 0]) / spacing).astype(int)
    ix, iy, iz = inds
    return (iz * nx * ny) + (iy * nx) + ix


def remap(x, out_min, out_max):
    in_min = np.min(x)
    in_max = np.max(x)
    return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min


def bandpass(data, band, sr, corners=4, zerophase=True):

    freqmin, freqmax = band
    fe = 0.5 * sr
    low = freqmin / fe
    high = freqmax / fe

    z, p, k = iirfilter(corners, [low, high], btype='band',
                        ftype='butter', output='zpk')
    sos = zpk2sos(z, p, k)
    if zerophase:
        firstpass = sosfilt(sos, data)
        if len(data.shape) == 1:
            return sosfilt(sos, firstpass[::-1])[::-1]
        else:
            return np.fliplr(sosfilt(sos, np.fliplr(firstpass)))
    else:
        return sosfilt(sos, data)


def filter(data, btype, band, sr, corners=4, zerophase=True):
    # btype: lowpass, highpass, band

    fe = 0.5 * sr
    z, p, k = iirfilter(corners, band / fe, btype=btype,
                        ftype='butter', output='zpk')
    sos = zpk2sos(z, p, k)
    if zerophase:
        firstpass = sosfilt(sos, data)
        if len(data.shape) == 1:
            return sosfilt(sos, firstpass[::-1])[::-1]
        else:
            return np.fliplr(sosfilt(sos, np.fliplr(firstpass)))
    else:
        return sosfilt(sos, data)


def decimate(data_in, sr, factor):
    data = data_in.copy()
    fmax = sr / (factor * 2)
    filter(data, 'lowpass', fmax, sr)
    if len(data.shape) == 1:
        return data[::factor]
    else:
        return data[:, ::factor]


def norm2d(d):
    return d / np.max(np.abs(d), axis=1)[:, np.newaxis]


def nextpow2(val):
    buf = math.ceil(math.log(val) / math.log(2))
    return int(math.pow(2, buf))


def cross_corr(sig1, sig2, norm=True, pad=False, phase_only=False, phat=False):
    """Cross-correlate two signals."""
    pad_len = len(sig1)
    if pad is True:
        pad_len *= 2
        # pad_len = signal.next_pow_2(pad_len)

    sig1f = fft(sig1, pad_len)
    sig2f = fft(sig2, pad_len)

    if phase_only is True:
        ccf = np.exp(- 1j * np.angle(sig1f)) * np.exp(1j * np.angle(sig2f))
    else:
        ccf = np.conj(sig1f) * sig2f

    if phat:
        ccf = ccf / np.abs(ccf)

    cc = np.real(ifft(ccf))

    if norm:
        cc /= np.sqrt(energy(sig1) * energy(sig2))

    return np.roll(cc, len(cc) // 2)


def unique_pairs(keys):
    return np.array(list(itertools.combinations(keys, 2)))


def ckeys_remove_intersta(ckeys, names):
    ikeep = []
    for i, ck in enumerate(names[ckeys]):
        n1, n2 = ck
        sta1 = ck[0].split('.')[0]
        sta2 = ck[1].split('.')[0]
        if sta1 == sta2:
            continue
        else:
            ikeep.append(i)

    return ckeys[np.array(ikeep)]


def ckeys_remove_intersta_str(ckeys):
    ikeep = []
    for i, ck in enumerate(ckeys):
        n1, n2 = ck
        sta1 = ck[0].split('.')[0]
        sta2 = ck[1].split('.')[0]
        if sta1 == sta2:
            continue
        else:
            ikeep.append(i)

    return ckeys[np.array(ikeep)]


def pairs_with_autocorr(keys):
    return np.array(list(itertools.combinations_with_replacement(keys, 2)))


def combos_between(keys1, keys2):
    return np.array(list(itertools.product(keys1, keys2)))


def dist_diff_ckeys(ckeys, locs):
    return np.linalg.norm(np.diff(locs[ckeys], axis=1).reshape(-1, 3), axis=1)


def xcorr_ckeys(dat, ckeys, norm=True):

    ncc = len(ckeys)
    nsta, wlen = dat.shape
    padlen = wlen * 2
    # nfreq = int(padlen // 2 + 1)
    # fstack = np.zeros((ncc, nfreq), dtype=np.complex64)
    stack = np.zeros((ncc, padlen), dtype=np.float32)

    fdat = np.fft.rfft(dat, axis=1, n=padlen)
    if norm:
        for irow in range(fdat.shape[0]):
            fdat[irow] /= np.sqrt(energy_freq(fdat[irow]))
            # fdat[irow] = filt * xutil.phase(fdat[irow])

    for j, ckey in enumerate(ckeys):
        k1, k2 = ckey
        stack[j] = np.fft.irfft(np.conj(fdat[k1]) * fdat[k2])

    stack = np.roll(stack, padlen // 2, axis=1)

    return stack


def keeplag(dat, nkeep):
    nsamp = dat.shape[-1]
    hl = int(nsamp // 2)
    return dat[:, hl - nkeep:hl + nkeep]


def energy(sig, axis=None):
    return np.sum(sig ** 2, axis=axis)


def energy_freq(fsig, axis=None):
    return np.sum(np.abs(fsig) ** 2, axis=axis) / fsig.shape[-1]


def norm_energy_freq(fsig):
    val = np.sqrt(np.sum(np.abs(fsig) ** 2) / len(fsig))
    return fsig / val


def build_slice_inds(start, stop, wlen, stepsize=None):

    if stepsize is None:
        stepsize = wlen

    overlap = wlen - stepsize
    imin = np.arange(start, stop - overlap - 1, stepsize)
    imax = np.arange(start + wlen, stop + stepsize - 1, stepsize)
    slices = np.dstack((imin, imax))[0].astype(int)
    if slices[-1][1] > stop:
        slices = slices[:-1]

    return slices


def build_welch_wins(start, stop, wlen, stepsize=None):

    if stepsize is None:
        stepsize = wlen

    overlap = wlen - stepsize
    imin = np.arange(start, stop - overlap - 1, stepsize)
    imax = np.arange(start + wlen, stop + stepsize - 1, stepsize)
    imid = imin + (imax - imin) // 2
    slices = np.dstack((imin, imid, imax))[0].astype(int)
    if slices[-1][1] > stop:
        slices = slices[:-1]

    return slices


def freq_window(cf, npts, sr, norm_energy=True):
    nfreq = int(npts // 2 + 1)
    fsr = npts / sr
    cf = np.array(cf, dtype=float)
    cx = (cf * fsr + 0.5).astype(int)

    win = np.zeros(nfreq, dtype=np.float32)
    win[:cx[0]] = 0
    win[cx[0]:cx[1]] = taper_cosine(cx[1] - cx[0])
    win[cx[1]:cx[2]] = 1
    win[cx[2]:cx[3]] = taper_cosine(cx[3] - cx[2])[::-1]
    win[cx[-1]:] = 0

    if norm_energy is True:
        win /= np.sqrt(energy_freq(win))

    return win


def taper_window(npts, taper_percentage):

    taplen = int(taper_percentage * npts)
    taper = taper_cosine(taplen)

    window = np.ones(npts, dtype=np.float32)
    window[:taplen] *= taper
    window[-taplen:] *= taper[::-1]
    return window


def taper_cosine(wlen):
    return np.cos(np.linspace(np.pi / 2., np.pi, wlen)) ** 2


def phase(sig):
    return np.exp(1j * np.angle(sig))


def whiten2D(a, freqs, sr):
    wl = a.shape[1]
    win = freq_window(freqs, wl, sr)
    af = fft(a)
    for sx in range(a.shape[0]):
        whiten_freq(af[sx], win)
    a[:] = np.real(ifft(af))


def whiten(sig, win):
    """Whiten signal, modified from MSNoise."""
    npts = len(sig)
    nfreq = int(npts // 2 + 1)

    assert(len(win) == nfreq)
    # fsr = npts / sr

    fsig = fft(sig)
    # hl = nfreq // 2

    half = fsig[: nfreq]
    half = win * phase(half)
    fsig[: nfreq] = half
    fsig[-nfreq + 1:] = half[1:].conjugate()[::-1]

    return np.real(ifft(fsig))


def whiten_freq(fsig, win):
    # npts = len(fsig)
    # nfreq = int(npts // 2 + 1)
    nfreq = int(len(fsig) // 2 + 1)
    assert(len(win) == nfreq)
    fsig[: nfreq] = win * phase(fsig[: nfreq])
    fsig[-nfreq + 1:] = fsig[1: nfreq].conjugate()[::-1]


def mirror_freqs(data):
    nfreq = int(data.shape[1] // 2 + 1)
    data[:, -nfreq + 1:] = np.fliplr(data[:, 1: nfreq].conjugate())


def taper_data(data, wlen):
    # tap = taper_cosine(wlen)
    tap = hann_half(wlen)
    data[:wlen] *= tap
    data[-wlen:] *= tap[::-1]


def taper2d(data, wlen):
    out = data.copy()
    tap = hann_half(wlen)
    for i in range(data.shape[0]):
        out[i][:wlen] *= tap
        out[i][-wlen:] *= tap[::-1]
    return out


def amax_cc(sig):
    return np.argmax(sig) - len(sig) // 2


def angle(a, b):
    # return np.arctan((a[1] - b[1]) / (a[0] - b[0]))
    return np.arctan2((a[1] - b[1]), (a[0] - b[0]))


def noise1d(wlen, freqs, sr, scale, taplen=0.05):

    # padlen =
    # fwin = freq_window(freqs, wlen, sr)
    # nfreq = int(npts // 2 + 1)
    # fwin = whiten_window(freqs, nfreq, sr)
    fwin = freq_window(freqs, wlen, sr)
    nfreq = len(fwin)

    fb = np.zeros(nfreq, dtype=np.complex64)
    phases = np.random.rand(nfreq) * 2 * np.pi
    phases = np.cos(phases) + 1j * np.sin(phases)

    fb = phases * fwin
    # out = np.real(np.fft.irfft(fb, n=nfreq)[:wlen]) * scale
    out = np.real(np.fft.irfft(fb, n=wlen)[:wlen]) * scale
    if taplen > 0:
        taper_data(out, int(taplen * wlen))

    return out


# def whiten_window(corners, nfreq, sr, norm_energy=True):

#     # freqs = np.fft.rfftfreq(nfreq, 1.0 / sr)
#     fsr = nfreq / sr
#     cf = np.array(corners, dtype=np.float32)
#     cx = (cf * fsr + 0.5).astype(int)
#     # print(cx)
#     # print(freqs[cx])

#     win = np.zeros(nfreq, dtype=np.float32)
#     win[:cx[0]] = 0
#     win[cx[0]:cx[1]] = taper_cosine(cx[1] - cx[0])
#     win[cx[1]:cx[2]] = 1
#     win[cx[2]:cx[3]] = taper_cosine(cx[3] - cx[2])[::-1]
#     win[cx[-1]:] = 0

#     if norm_energy is True:
#         win /= np.sqrt(energy_freq(win))

#     return win


# def noise1d(npts, freqs, sr, scale, taplen=0.05):

#     out = np.zeros(npts, dtype=np.float32)
#     fwin = freq_window(freqs, npts, sr)
#     nfreq = len(fwin)

#     fb = np.zeros(npts, dtype=np.complex64)
#     fb = np.zeros(npts, dtype=np.complex64)
#     phases = np.random.rand(nfreq) * 2 * np.pi
#     phases = np.cos(phases) + 1j * np.sin(phases)

#     fb[: nfreq] = phases * fwin
#     fb[-nfreq + 1:] = fb[1:nfreq].conjugate()[::-1]
#     # a[i] = np.real(ifft(fb))
#     out += np.real(ifft(fb)) * scale
#     if taplen > 0:
#         taper_data(out, int(taplen * npts))

#     return out


def add_noise(a, freqs, sr, scale, taplen=0.05):
    out = a.copy()

    nsig, npts = a.shape
    fwin = freq_window(freqs, npts, sr)
    nfreq = len(fwin)

    for i in range(nsig):
        fb = np.zeros(npts, dtype=np.complex64)

        phases = np.random.rand(nfreq) * 2 * np.pi
        phases = np.cos(phases) + 1j * np.sin(phases)

        fb[: nfreq] = phases * fwin
        fb[-nfreq + 1:] = fb[1:nfreq].conjugate()[::-1]
        # a[i] = np.real(ifft(fb))
        out[i] += np.real(ifft(fb)) * scale
    taper2d(out, int(taplen * npts))

    return out


def zeropad2d(a, npad):
    nrow, ncol = a.shape
    out = np.zeros((nrow, ncol + npad), dtype=a.dtype)
    out[:, :ncol] = a
    return out


def zero_argmax(a, wlen, taplen=0.05):
    # handles edges incorrectly
    npts = a.shape[1]
    taper = hann_half(int(taplen * wlen))
    win = np.concatenate((taper[::-1], np.zeros(wlen), taper))
    hl = int(len(win) // 2)
    out = a.copy()
    imaxes = np.argmax(np.abs(out), axis=1)
    for i, imax in enumerate(imaxes):
        i0 = imax - hl
        i1 = imax + hl
        if i0 <= 0:
            out[i, 0:i1] *= win[abs(i0):]
        elif i1 > npts:
            out[i, i0:] *= win[:-(i1 - npts)]
        else:
            out[i, i0:i1] *= win

    return out


def hann(npts):
    return 0.5 - 0.5 * np.cos(2.0 * np.pi * np.arange(npts) / (npts - 1))


def hann_half(npts):
    return hann(npts * 2)[:npts]

# def whiten(sig, win):
#   """Whiten signal, modified from MSNoise."""
#   npts = len(sig)
#   nfreq = int(npts // 2 + 1)

#   assert(len(win) == nfreq)
#   # fsr = npts / sr

#   fsig = fft(sig)
#   # hl = nfreq // 2

#   half = fsig[: nfreq]
#   half = win * phase(half)
#   fsig[: nfreq] = half
#   fsig[-nfreq + 1:] = half[1:].conjugate()[::-1]

#   return np.real(ifft(fsig))


def fftnoise(f):
    f = np.array(f, dtype='complex')
    npts = (len(f) - 1) // 2
    phases = np.random.rand(npts) * 2 * np.pi
    phases = np.cos(phases) + 1j * np.sin(phases)
    f[1:npts + 1] *= phases
    f[-1:-1 - npts:-1] = np.conj(f[1:npts + 1])
    return ifft(f).real


def band_noise(band, sr, samples):
    freqmin, freqmax = band
    freqs = np.abs(fftfreq(samples, 1. / sr))
    f = np.zeros(samples)
    idx = np.where(np.logical_and(freqs >= freqmin, freqs <= freqmax))[0]
    f[idx] = 1
    return fftnoise(f)


def add_noise2(data, band, sr, power):
    freqmin, freqmax = band
    samples = data.shape[1]
    for d in data:
        d += band_noise(band, sr, samples) * power


# def envelope(data):
#     slen = len(data)
#     FFT = fft(data, slen)
#     FFT[1: slen // 2] *= 2
#     FFT[slen // 2:] = 0
#     return np.abs(ifft(FFT))

def envelope(data, **kwargs):
    return np.abs(hilbert(data, **kwargs))


def smooth(data, nsmooth, **kwargs):
    return gaussian_filter(data, nsmooth, **kwargs)

# def pairs_excluding_acoustic():
        # fd = rfft(sig)
    # # keys = np.arange(ds.shape[0])
    # # ckeys = np.array(list(itertools.combinations(keys, 2))).astype(np.int32)
    # ckeys = ck
    # scut = 700
    # avel = 345.
    # sloc = srcs[0]

    # dd0 = np.array([xutil.dist_between(l, sloc) for l in locs])
    # dd1 = np.diff(dd0[ckeys], axis=1)
    # ia = dd1 / avel * sr
    # # plt.hist(np.abs(ia), bins=100)
    # ik0 = np.where(np.abs(ia) < scut)[0]
    # sloc = srcs[1]
    # dd0 = np.array([xutil.dist_between(l, sloc) for l in locs])
    # dd1 = np.diff(dd0[ckeys], axis=1)
    # ia = dd1 / avel * sr
    # # plt.hist(np.abs(ia), bins=100)
    # ik1 = np.where(np.abs(ia) < scut)[0]
    # ik = np.unique(np.concatenate((ik0, ik1)))
    # ck2 = np.delete(ckeys, ik, axis=0)
    # xplot.stations(locs, ckeys=ck2[::500])
    # dd = np.linalg.norm(np.diff(locs[ck2], axis=1).reshape(-1, 3), axis=1)
    # # plt.hist(dd, bins=100)
    # cki = np.where(dd < 1000)[0]

def build_checkerboard(w, h):
    re = np.r_[w * [0, 1]]              # even-numbered rows
    ro = np.r_[w * [1, 0]]              # odd-numbered rows
    return np.row_stack(h * (re, ro))


def nan_helper(y):
    """Helper to handle indices and logical indices of NaNs.

    Input:
        - y, 1d numpy array with possible NaNs
    Output:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices
    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    """

    return np.isnan(y), lambda z: z.nonzero()[0]


def nans_interp(data, threshold=.05):
    nans, x = nan_helper(data)
    number_of_nans = data[nans].size
    if number_of_nans > 0:
        if float(number_of_nans) / len(data) < threshold:
            data[nans] = np.interp(x(nans), x(~nans), data[~nans])
        else:
            data[:] = 0


# def xcorr_pair_stack_slices(sig1, sig2, cclen, stacklen, keeplag, stepsize=None, whiten_freqs=None, sr=None, onebit=False):

#     cclen = int(cclen)
#     keeplag = int(keeplag)
#     slices = build_slice_inds(0, len(sig1), cclen, stepsize=stepsize)
#     xvals = np.mean(slices, axis=1)

#     ncc = int(slices[-1][1] / stacklen)
#     padlen = cclen * 2
#     nfreq = int(padlen // 2 + 1)
#     # ccs = np.zeros((ncc, padlen), dtype=np.float32)
#     keeplen = int(keeplag * 2)
#     ccs = np.zeros((ncc, keeplen), dtype=np.float32)

#     win = None
#     if whiten_freqs is not None:
#         win = freq_window(whiten_freqs, padlen, sr)

#     step = slices[1][0] - slices[0][0]
#     assert(stacklen % step == 0)
#     chunksize = stacklen // step
#     print(step, chunksize)
#     print(ncc)

#     fcc = np.zeros(nfreq, dtype=np.complex64)
#     tmp_nstack = 0
#     cix = 0

#     for i, sl in enumerate(slices):
#         print(f"{i} / {len(slices)}")
#         win1 = sig1[sl[0]:sl[1]]
#         win2 = sig2[sl[0]:sl[1]]

#         if onebit is True:
#             win1 = np.sign(win1)
#             win2 = np.sign(win2)

#         f1 = np.fft.rfft(win1, n=padlen)
#         f2 = np.fft.rfft(win2, n=padlen)

#         if win is not None:
#             f1 = win * phase(f1)
#             f2 = win * phase(f2)

#         f1 /= np.sqrt(energy_freq(f1))
#         f2 /= np.sqrt(energy_freq(f2))
#         fcc += np.conj(f1) * f2
#         tmp_nstack += 1

#         if tmp_nstack >= chunksize:
#             cc = np.fft.irfft(fcc) / tmp_nstack
#             ccs[cix, :keeplag] = cc[-keeplag:]
#             ccs[cix, keeplag:] = cc[:keeplag]

#             fcc.fill(0)
#             tmp_nstack = 0
#             cix += 1

#     return xvals, ccs


# def xcorr_ckeys_stack_slices(rawdat, ckeys, cclen, keeplag, stepsize=None, whiten_freqs=None, sr=None, onebit=False):

#     cclen = int(cclen)

#     nchan, nsamp = rawdat.shape
#     slices = build_slice_inds(0, nsamp, cclen, stepsize=stepsize)
#     # xvals = np.mean(slices, axis=1)

#     # ncc = int(slices[-1][1] / stacklen)
#     ncc = len(ckeys)
#     padlen = cclen * 2
#     nfreq = int(padlen // 2 + 1)

#     whiten_win = None
#     if whiten_freqs is not None:
#         whiten_win = freq_window(whiten_freqs, padlen, sr)

#     fstack = np.zeros((ncc, nfreq), dtype=np.complex64)

#     for i, sl in enumerate(slices):
#         # print(f"{i} / {len(slices)}")
#         dat = rawdat[:, sl[0]:sl[1]].copy()

#         if onebit is True:
#             dat = bandpass(dat, whiten_freqs[[0, -1]], sr)
#             dat = np.sign(dat)

#         fdat = np.fft.rfft(dat, axis=1, n=padlen)

#         if whiten_win is not None:
#             for irow in range(fdat.shape[0]):
#                 fdat[irow] = whiten_win * phase(fdat[irow])
#         else:
#             for irow in range(fdat.shape[0]):
#                 fdat[irow] /= np.sqrt(energy_freq(fdat[irow]))

#         # dat0 = np.fft.irfft(fdat, axis=1)
#         # norm = xutil.energy_freq(fdat[0])

#         for j, ckey in enumerate(ckeys):
#             k1, k2 = ckey
#             # stack[j] = np.fft.irfft(np.conj(fdat[k1]) * fdat[k2])
#             fstack[j] += np.conj(fdat[k1]) * fdat[k2]

#     # stack = np.fft.irfft(fstack, axis=1)
#     # stack = np.roll(stack, padlen // 2, axis=1)
#     keeplag = int(keeplag)
#     stack = np.zeros((ncc, keeplag * 2), dtype=np.float32)

#     for i in range(ncc):
#         cc = np.fft.irfft(fstack[i]) / len(slices)
#         stack[i, :keeplag] = cc[-keeplag:]
#         stack[i, keeplag:] = cc[:keeplag]

#     return stack


# def xcorr_freq(sig1f, sig2f):
#   """Cross-correlate two signals."""

#   ccf = np.conj(sig1f) * sig2f

#   if phat:
#       ccf = ccf / np.abs(ccf)

#   cc = np.real(ifft(ccf))

#   if norm:
#       cc /= np.sqrt(energy(sig1) * energy(sig2))

#   return np.roll(cc, len(cc) // 2)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xseis2']

package_data = \
{'': ['*'], 'xseis2': ['include/gsl/*', 'include/xseis2/*']}

install_requires = \
['Cython>=0.29.25,<0.30.0',
 'conan>=1.43.0,<2.0.0',
 'obspy>=1.2.2,<2.0.0',
 'setuptools>=59.6.0,<60.0.0']

setup_kwargs = {
    'name': 'xseis2',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# distutils: language = c++
# cython: language_level=3
# Cython interface file for wrapping the object
cimport numpy as np
import numpy as np
from libcpp cimport bool
from libcpp.string cimport string
from libc.stdint cimport uint16_t, uint32_t, int64_t
from libc.stdint cimport uintptr_t
# from libc.stdint cimport uintptr_t
from libcpp.vector cimport vector

ctypedef uint16_t* uint16_ptr # pointer workaround

cdef extern from "xseis2/workflows.h" namespace "xs":

	void SearchOnePhase(
		float* raw_data_ptr, uint32_t nchan, uint32_t nsample, # ptr to raw data (size = nchan x nsamp)
		float samplerate,
		float* station_locations_ptr, uint32_t nsta, # ptr to station locs (size = nsta x 3)
		uint16_t* channel_map_ptr, # ptr to channel map (size = nchan)
		vector[uint16_ptr]& ttable_row_ptrs, uint32_t ngrid,
		int64_t* outbuf,
		vector[float]& whiten_corner_freqs, # 4 corner frequencies to whiten between
		float pair_dist_min, # uses only correlation pairs between dist_min and dist_max
		float pair_dist_max,
		float cc_smooth_length_sec, # length to absolute value smooth ccfs before beamforming
		uint32_t nthreads,
		int debug_lvl,
		string& debug_file
		)


def pySearchOnePhase(np.ndarray[np.float32_t, ndim=2] data,
					samplerate,
					np.ndarray[np.uint16_t, ndim=1] chanmap,
					np.ndarray[np.float32_t, ndim=2] stalocs,
					np.ndarray[np.int64_t, ndim=1] tt_ptrs,
					ngrid,
					np.ndarray[np.float32_t, ndim=1] whiten_corner_freqs,
					pair_dist_min,
					pair_dist_max,
					cc_smooth_length_sec,
					nthreads,
					debug_lvl,
					debug_file
			   ):

# Pre-process, cross-correlate and beamform raw_data using provided traveltime table.
# The provided traveltimes determine which phase (e.g P,S) is searched for.

# :param data: 2d array of channel data (1 row per channel), assumes all starttimes equal
# :param samplerate: sampling rate of data
# :param chanmap: indexes to map each channel to its respective station location
# :param stalocs: unique station locations (shape = nsta x 3)
# :param tt_ptrs: pointers to rows in the traveltime table corresponding to stalocs
# :param ngrid: number of columns (or search locations) in the traveltime tables
# :param whiten_corner_freqs: array of 4 floats defining whitening corner frequencies in Hz
# :param pair_dist_min: uses only correlation pairs between dist_min and dist_max (meters)
# :param pair_dist_max:
# :param cc_smooth_length_sec: length to absolute value smooth ccfs before beamforming
# :param debug_lvl: int 0 (disabled) 1 (logging) 2 (logging + data dump) to debug_file


# * Important that rows of tt_ptrs are consistent with rows of station locations
# * The chanmap is used to determine the traveltimes to use for each channel
#   in the data array.


	chanmap = np.ascontiguousarray(chanmap)
	stalocs = np.ascontiguousarray(stalocs)
	data = np.ascontiguousarray(data)

	assert(data.shape[0] == chanmap.shape[0])
	assert(tt_ptrs.shape[0] == stalocs.shape[0])

	# convert np array to std::vector
	cdef vector[float] corner_freqs_vec
	for i in range(whiten_corner_freqs.shape[0]):
		corner_freqs_vec.push_back(<float>whiten_corner_freqs[i])

	# buil vec of c-pointers from casted python ptrs
	cdef vector[uint16_ptr] tt_ptrs_vec
	for i in range(tt_ptrs.shape[0]):
		tt_ptrs_vec.push_back(<uint16_t*>tt_ptrs[i])

	cdef string debug_file_str = debug_file.encode('UTF-8')
	cdef np.ndarray[np.int64_t, ndim=1] out = np.empty(3, dtype=np.int64)
	# print("{0:x}".format(<unsigned long>&stalocs[0, 0]))

	SearchOnePhase(&data[0, 0], data.shape[0], data.shape[1],
					samplerate,
					&stalocs[0, 0], stalocs.shape[0],
					&chanmap[0],
					tt_ptrs_vec, ngrid,
					&out[0],
					corner_freqs_vec,
					pair_dist_min,
					pair_dist_max,
					cc_smooth_length_sec,
					nthreads,
					debug_lvl,
					debug_file_str
					)

	out_list = [float(out[0]) / 10000., out[1], out[2]]
	return out_list


